import numpy as np
from skimage.transform import resize


def transform_spatial(coordinates, original, resized):
    # obs transform
    x = coordinates[:, 0]
    y = coordinates[:, 1]
    x1p = x / original[1]
    y1p = y / original[0]
    x1 = x1p * resized[1]
    y1 = y1p * resized[0]

    return np.vstack([x1, y1]).transpose()


def correct_size(adata, fixed_size):

    image = adata.uns["spatial"][list(adata.uns["spatial"].keys())[0]]["images"][
        "hires"
    ]
    image_size = image.shape[:2]
    if image_size != fixed_size:
        adata.obs[["imagerow", "imagecol"]] = transform_spatial(
            adata.obs[["imagerow", "imagecol"]].values, image_size, fixed_size
        )
        adata.obsm["spatial"] = transform_spatial(
            adata.obsm["spatial"], image_size, fixed_size
        )
        image_resized = resize(image, fixed_size)
        adata.uns["spatial"][list(adata.uns["spatial"].keys())[0]]["images"][
            "hires"
        ] = image_resized

    return adata


def concatenate_spatial_adata(adata_list, ncols=2, fixed_size=(2000, 2000)):
    """\
    Concatnate multiple anndata for visualization of spatial transcriptomics

    Parameters
    ----------
    adata_list
        A list of anndata objet.
    ncols
        Number of columns
    fixed_size
        The size that fixed for every spatial transcriptomics data
    Returns
    -------
    Returns adata.
    """

    use_adata_list = []
    for adata in adata_list:
        use_adata_list.append(adata.copy())

    import math

    # check valid
    n_adata = len(use_adata_list)
    nrows = math.ceil(n_adata / ncols)
    if ncols > n_adata:
        raise ValueError("Number of column is out of bound")

    # Correct size
    for adata in use_adata_list:
        correct_size(adata, fixed_size=fixed_size)

    # Transform
    n = 0
    break_out_flag = False
    scale = use_adata_list[0].uns["spatial"][
        list(use_adata_list[0].uns["spatial"].keys())[0]
    ]["scalefactors"]["tissue_hires_scalef"]
    for i in range(0, nrows):
        for j in range(0, ncols):
            obs_spatial = use_adata_list[n].obs[["imagerow", "imagecol"]].values
            obsm_spatial = use_adata_list[n].obsm["spatial"]
            obs_spatial = np.vstack(
                (
                    obs_spatial[:, 0] + fixed_size[0] * i,
                    obs_spatial[:, 1] + fixed_size[1] * j,
                )
            ).transpose()
            obsm_spatial = np.vstack(
                (
                    obsm_spatial[:, 0] + fixed_size[0] / scale * i,
                    obsm_spatial[:, 1] + fixed_size[1] / scale * j,
                )
            ).transpose()
            use_adata_list[n].obs[["imagerow", "imagecol"]] = obs_spatial
            use_adata_list[n].obsm["spatial"] = obsm_spatial
            if n == len(use_adata_list) - 1:
                break_out_flag = True
                break
            n += 1
        if break_out_flag:
            break

    # Combine images
    imgs = []
    for i, adata in enumerate(use_adata_list):
        imgs.append(
            adata.uns["spatial"][list(adata.uns["spatial"].keys())[0]]["images"][
                "hires"
            ]
        )

    from PIL import Image

    if (nrows * ncols - len(use_adata_list)) > 0:
        for i in range(0, (nrows * ncols - len(use_adata_list))):
            image = Image.new("RGB", fixed_size, (255, 255, 255, 255))
            imgs.append(np.array(image))

    print(len(imgs))

    img_rows = []
    for min_id in range(0, len(use_adata_list), ncols):
        img_row = np.hstack(imgs[min_id : min_id + ncols])
        img_rows.append(img_row)
    imgs_comb = np.vstack((i for i in img_rows))

    adata_concat = use_adata_list[0].concatenate(use_adata_list[1:])
    adata_concat.uns["spatial"] = use_adata_list[0].uns["spatial"]

    adata_concat.uns["spatial"][list(adata_concat.uns["spatial"].keys())[0]]["images"][
        "hires"
    ] = imgs_comb

    return adata_concat



import numpy as np
import pandas as pd

import io
from PIL import Image

import matplotlib
import matplotlib.pyplot as plt
from anndata import AnnData
from scanpy.plotting import palettes
from stlearn.plotting import palettes_st

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from enum import Enum

from matplotlib import rcParams, ticker, gridspec, axes
from matplotlib.axes import Axes
from abc import ABC


def get_img_from_fig(fig, dpi=180):
    buf = io.BytesIO()
    from io import BytesIO

    fig.savefig(
        buf, format="png", dpi=dpi, bbox_inches="tight", pad_inches=0, transparent=True
    )
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    img = np.asarray(Image.open(BytesIO(img_arr)))
    buf.close()
    # img = cv2.imdecode(img_arr, 1)
    # img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGB)

    return img


def centroidpython(x, y):
    l = len(x)
    return sum(x) / l, sum(y) / l


def get_cluster(search, dictionary):
    for (
        cl,
        sub,
    ) in (
        dictionary.items()
    ):  # for name, age in dictionary.iteritems():  (for Python 2.x)
        if search in sub:
            return cl


def get_node(node_list, split_node):
    result = np.array([])
    for node in node_list:
        result = np.append(result, np.array(split_node[node]).astype(int))
    return result.astype(int)


def check_sublist(full, sub):
    set_sub = set(sub)
    index_bool = [x in set_sub for x in full]

    return index_bool


def get_cmap(cmap):
    """Checks inputted cmap string."""
    if cmap == "vega_10_scanpy":
        cmap = palettes.vega_10_scanpy
    elif cmap == "vega_20_scanpy":
        cmap = palettes.vega_20_scanpy
    elif cmap == "default_102":
        cmap = palettes.default_102
    elif cmap == "default_28":
        cmap = palettes.default_28
    elif cmap == "jana_40":
        cmap = palettes_st.jana_40
    elif cmap == "default":
        cmap = palettes_st.default
    elif type(cmap) == str:  # If refers to matplotlib cmap
        cmap_n = plt.get_cmap(cmap).N
        return plt.get_cmap(cmap), cmap_n
    elif type(cmap) == matplotlib.colors.LinearSegmentedColormap:  # already cmap
        cmap_n = cmap.N
        return cmap, cmap_n

    cmap_n = len(cmap)
    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", cmap)

    cmap_ = plt.cm.get_cmap(cmaps)

    return cmap_, cmap_n


def check_cmap(cmap):
    """Initialize cmap"""
    scanpy_cmap = ["vega_10_scanpy", "vega_20_scanpy", "default_102", "default_28"]
    stlearn_cmap = ["jana_40", "default"]
    cmap_available = plt.colormaps() + scanpy_cmap + stlearn_cmap
    error_msg = (
        "cmap must be a matplotlib.colors.LinearSegmentedColormap OR"
        "one of these: " + str(cmap_available)
    )
    if type(cmap) == str:
        assert cmap in cmap_available, error_msg
    elif type(cmap) != matplotlib.colors.LinearSegmentedColormap:
        raise Exception(error_msg)

    return cmap


def get_colors(adata, obs_key, cmap="default", label_set=None):
    """Retrieves colors if present in adata.uns, if not present then will set
    them as per scanpy & return in order requested.
    """
    # Checking if colors are already set #
    col_key = f"{obs_key}_colors"
    if col_key in adata.uns:
        labels_ordered = adata.obs[obs_key].cat.categories
        colors_ordered = adata.uns[col_key]
    else:  # Colors not already present
        check_cmap(cmap)
        cmap, cmap_n = get_cmap(cmap)

        if not hasattr(adata.obs[obs_key], "cat"):  # Ensure categorical
            adata.obs[obs_key] = adata.obs[obs_key].astype("category")
        labels_ordered = adata.obs[obs_key].cat.categories
        colors_ordered = [
            matplotlib.colors.rgb2hex(cmap(i / (len(labels_ordered) - 1)))
            for i in range(len(labels_ordered))
        ]
        adata.uns[col_key] = colors_ordered

    # Returning the colors of the desired labels in indicated order #
    if type(label_set) != type(None):
        colors_ordered = [
            colors_ordered[np.where(labels_ordered == label)[0][0]]
            for label in label_set
        ]

    return colors_ordered



""" Wrapper function for performing CCI analysis, varrying the analysis based on
    the inputted data / state of the anndata object.
"""

import os
import numba
import numpy as np
import pandas as pd
from typing import Union
from anndata import AnnData
from .base import calc_neighbours, get_lrs_scores, calc_distance
from .permutation import perform_spot_testing
from .go import run_GO
from .het import (
    count,
    get_data_for_counting,
    get_interaction_matrix,
    get_interaction_pvals,
)
from statsmodels.stats.multitest import multipletests

################################################################################
# Functions related to Ligand-Receptor interactions #
################################################################################
def load_lrs(names: Union[str, list, None] = None, species: str = "human") -> np.array:
    """Loads inputted LR database, & concatenates into consistent database set of pairs without duplicates. If None loads 'connectomeDB2020_lit'.

    Parameters
    ----------
    names: list     Databases to load, options: 'connectomeDB2020_lit' (literature verified), 'connectomeDB2020_put' (putative). If more than one specified, loads all & removes duplicates.
    species: str    Format of the LR genes, either 'human' or 'mouse'.
    Returns
    -------
    lrs: np.array   lr pairs from the database in format ['L1_R1', 'LN_RN']
    """
    if type(names) == type(None):
        names = ["connectomeDB2020_lit"]
    if type(names) == str:
        names = [names]

    path = os.path.dirname(os.path.realpath(__file__))
    dbs = [pd.read_csv(f"{path}/databases/{name}.txt", sep="\t") for name in names]
    lrs_full = []
    for db in dbs:
        lrs = [f"{db.values[i,0]}_{db.values[i,1]}" for i in range(db.shape[0])]
        lrs_full.extend(lrs)
    lrs_full = np.unique(lrs_full)
    # If dealing with mouse, need to reformat #
    if species == "mouse":
        genes1 = [lr_.split("_")[0] for lr_ in lrs_full]
        genes2 = [lr_.split("_")[1] for lr_ in lrs_full]
        lrs_full = np.array(
            [
                genes1[i][0]
                + genes1[i][1:].lower()
                + "_"
                + genes2[i][0]
                + genes2[i][1:].lower()
                for i in range(len(lrs_full))
            ]
        )

    return lrs_full


def grid(adata, n_row: int = 10, n_col: int = 10, use_label: str = None):
    """Creates a new anndata representing a gridded version of the data; can be
        used upstream of CCI pipeline. NOTE: intended use is for single cell
        spatial data, not Visium or other lower resolution tech.

    Parameters
    ----------
    adata: AnnData
        The data object.
    n_row: int
        The number of rows in the grid.
    n_col: int
        The number of columns in the grid.
    use_label: str
        The cell type labels in adata.obs to join together & save as deconvolution data.
    Returns
    -------
    grid_data: AnnData
        Equivalent expression data to adata, except values have been summed by
        cells that fall within defined bins.
    """

    # Retrieving the coordinates of each grid #
    n_squares = n_row * n_col
    cell_bcs = adata.obs_names.values
    xs, ys = adata.obs["imagecol"].values, adata.obs["imagerow"].values

    grid_counts, xedges, yedges = np.histogram2d(xs, ys, bins=[n_col, n_row])

    grid_expr = np.zeros((n_squares, adata.shape[1]))
    grid_coords = np.zeros((n_squares, 2))
    grid_bcs = []
    grid_cell_counts = []
    gridded_cells = []
    cell_grid = []
    # If use_label specified, then will generate deconvolution information
    if type(use_label) != type(None):
        cell_labels = adata.obs[use_label].values.astype(str)
        cell_set = np.unique(cell_labels)
        cell_info = np.zeros((n_squares, len(cell_set)))

    # generate grids from top to bottom and left to right
    n = 0
    for i in range(n_col):
        x_left, x_right = xedges[i], xedges[i + 1]
        for j in range(n_row):
            y_down, y_up = yedges[j], yedges[j + 1]
            grid_coords[n, :] = [(x_right + x_left) / 2, (y_up + y_down) / 2]

            # Now determining the cells within the gridded area #
            if i != n_col - 1 and j == n_row - 1:  # top left corner
                x_true = (xs >= x_left) & (xs < x_right)
                y_true = (ys <= y_up) & (ys > y_down)
            elif i == n_col - 1 and j != n_row - 1:  # bottom righ corner
                x_true = (xs > x_left) & (xs <= x_right)
                y_true = (ys < y_up) & (ys >= y_down)
            else:  # average case
                x_true = (xs >= x_left) & (xs < x_right)
                y_true = (ys < y_up) & (ys >= y_down)

            grid_cells = cell_bcs[x_true & y_true]
            grid_cells_str = ",".join(grid_cells)
            grid_bcs.append(grid_cells_str)
            grid_cell_counts.append(len(grid_cells))
            gridded_cells.extend(grid_cells)
            cell_grid.extend([f"grid_{n}"] * len(grid_cells))

            # Summing the expression across these cells to get the grid expression #
            if len(grid_cells) > 0:
                cell_bool = [cell in grid_cells for cell in cell_bcs]
                grid_expr[n, :] = adata.X[cell_bool, :].sum(axis=0)

            # If we have cell type information, will record #
            if type(use_label) != type(None) and len(grid_cells) > 0:
                grid_cell_types = cell_labels[cell_bool]
                cell_info[n, :] = [
                    len(np.where(grid_cell_types == ct)[0]) / len(grid_cell_types)
                    for ct in cell_set
                ]

            n += 1

    # Creating gridded anndata #
    grid_expr = pd.DataFrame(
        grid_expr,
        index=[f"grid_{i}" for i in range(n_squares)],
        columns=adata.var_names.values.astype(str),
    )
    grid_data = AnnData(grid_expr)
    grid_data.obs["imagecol"] = grid_coords[:, 0]
    grid_data.obs["imagerow"] = grid_coords[:, 1]
    grid_data.obs["n_cells"] = grid_cell_counts
    grid_data.obsm["spatial"] = grid_coords
    grid_data.uns["spatial"] = adata.uns["spatial"]

    if type(use_label) != type(None):
        grid_data.uns[use_label] = pd.DataFrame(
            cell_info, index=grid_data.obs_names.values.astype(str), columns=cell_set
        )
        max_indices = np.apply_along_axis(np.argmax, 1, cell_info)
        cell_set = np.unique(grid_data.uns[use_label].columns.values)
        grid_data.obs[use_label] = [cell_set[index] for index in max_indices]
        grid_data.obs[use_label] = grid_data.obs[use_label].astype("category")

    # Subsetting to only gridded spots that contain cells #
    grid_data = grid_data[grid_data.obs["n_cells"] > 0, :].copy()
    if type(use_label) != type(None):
        grid_data.uns[use_label] = grid_data.uns[use_label].loc[grid_data.obs_names, :]

    grid_data.uns["grid_counts"] = grid_counts
    grid_data.uns["grid_xedges"] = xedges
    grid_data.uns["grid_yedges"] = yedges

    return grid_data


def run(
    adata: AnnData,
    lrs: np.array,
    min_spots: int = 10,
    distance: int = None,
    n_pairs: int = 1000,
    n_cpus: int = None,
    use_label: str = None,
    adj_method: str = "fdr_bh",
    pval_adj_cutoff: float = 0.05,
    min_expr: float = 0,
    save_bg: bool = False,
    neg_binom: bool = False,
    verbose: bool = True,
):
    """Performs stLearn LR analysis.

    Parameters
    -----------
    adata: AnnData
        The data object.
    lrs: np.array
        The LR pairs to score/test for enrichment (in format 'L1_R1').
    min_spots: int
        Minimum number of spots with an LR score for an LR to be considered for
        further testing.
    distance: int
        Distance to determine the neighbours (default [None] is immediately
        adjacent neighbours if using Visium), distance=0 means within spot
        (only for non-single-cell spatial data).
    n_pairs: int
        Number of random pairs of genes to generate when creating the background
        distribution per LR pair; higher than more accurate p-value estimation.
    n_cpus: int
        The number of cpus to use for multi-threading.
    use_label: str
        The cell type deconvolution results to use in counting stored in
        adata.uns; if not specified only considered LR expression without cell
        heterogeneity.
    adj_method: str
        Parsed to statsmodels.stats.multitest.multipletests for multiple
        hypothesis testing correction; see there for other options.
    pval_adj_cutoff: float
        P-value below which LR is considered significant in spot neighbourhood.
    min_expr: float
        Minimum gene expression of either L or R for spot to be considered to
        expression of either.
    save_bg: bool
        Whether to save the background per LR pair; for method development only.
        Not recommended since huge memory.
    neg_binom: bool
        Whether to fit a negative binomial distribution for all background
        scores generated across spots per LR after discretising the random
        scores. Can be extremely slow.
    verbose: bool
        Whether print dialogue to user during run-time.
    Returns
    --------
    adata: AnnData
    Relevant information stored:
        adata.uns['lr_summary']
            Summary of significant spots detected per LR,
            the LRs listed in the index is the same order of LRs in the columns of
            results stored in adata.obsm below. Hence the order of this must be maintained.
        adata.obsm
            Additional keys are added; 'lr_scores', 'lr_sig_scores', 'p_vals',
            'p_adjs', '-log10(p_adjs)'. All are numpy matrices, with columns
            referring to the LRs listed in adata.uns['lr_summary']. 'lr_scores'
            is the raw scores, while 'lr_sig_scores' is the same except only for
            significant scores; non-significant scores are set to zero.
        adata.obsm['het']
            Only if use_label specified; contains the counts of the cell types found per spot.
    """
    # Setting threads for paralellisation #
    if type(n_cpus) != type(None):
        numba.set_num_threads(n_cpus)

    # Making sure none of the var_names contains '_' already, these will need
    # to be renamed.
    prob_genes = [gene for gene in adata.var_names if "_" in gene]
    if len(prob_genes) > 0:
        raise Exception(
            "Detected '_' within some gene names, which breaks "
            + "internal string handling for the lrs in format 'L_R'.\n"
            + "Recommend to rename adata.var_names or remove these "
            + f"genes from adata:\n {prob_genes}"
        )

    # Calculating neighbour & storing #
    distance = calc_distance(adata, distance)
    neighbours = calc_neighbours(adata, distance, verbose=verbose)
    adata.obsm["spot_neighbours"] = pd.DataFrame(
        [",".join(x.astype(str)) for x in neighbours],
        index=adata.obs_names,
        columns=["neighbour_indices"],
    )
    spot_neighs_df = adata.obsm["spot_neighbours"]
    spot_neigh_bcs = []
    for i in range(spot_neighs_df.shape[0]):
        neigh_indices = [
            int(index)
            for index in spot_neighs_df.values[i, 0].split(",")
            if index != ""
        ]
        neigh_bcs = [adata.obs_names[index] for index in neigh_indices]
        spot_neigh_bcs.append(",".join(neigh_bcs))
    spot_neigh_bcs_df = pd.DataFrame(
        spot_neigh_bcs, index=spot_neighs_df.index, columns=["neighbour_bcs"]
    )
    # Important to store barcodes in-case adata subsetted #
    adata.obsm["spot_neigh_bcs"] = spot_neigh_bcs_df

    if verbose:
        print(
            "Spot neighbour indices stored in adata.obsm['spot_neighbours'] "
            "& adata.obsm['spot_neigh_bcs']."
        )

    # Conduct with cell heterogeneity info if label_transfer provided #
    cell_het = type(use_label) != type(None) and use_label in adata.uns.keys()
    if cell_het:
        if verbose:
            print("Calculating cell hetereogeneity...")

        # Calculating cell heterogeneity #
        count(adata, distance=distance, use_label=use_label, use_het=use_label)

    het_vals = (
        np.array([1] * len(adata))
        if use_label not in adata.obsm
        else adata.obsm[use_label]
    )

    """ 1. Filter any LRs without stored expression.
    """
    # Calculating the lr_scores across spots for the inputted lrs #
    lr_scores, lrs = get_lrs_scores(adata, lrs, neighbours, het_vals, min_expr)
    lr_bool = (lr_scores > 0).sum(axis=0) > min_spots
    lrs = lrs[lr_bool]
    lr_scores = lr_scores[:, lr_bool]
    if verbose:
        print("Altogether " + str(len(lrs)) + " valid L-R pairs")
    if len(lrs) == 0:
        print("Exiting due to lack of valid LR pairs.")
        return

    """ Permutation methods generating background per spot, & test lrs in spot.
    """
    perform_spot_testing(
        adata,
        lr_scores,
        lrs,
        n_pairs,
        neighbours,
        het_vals,
        min_expr,
        adj_method,
        pval_adj_cutoff,
        verbose,
        save_bg=save_bg,
        neg_binom=neg_binom,
    )


def adj_pvals(
    adata,
    pval_adj_cutoff: float = 0.05,
    correct_axis: str = "spot",
    adj_method: str = "fdr_bh",
):
    """Performs p-value adjustment and determination of significant spots.
        Default settings of this function are already run in st.tl.cci.run.

    Parameters
    ----------
    adata: AnnData
        Must have st.tl.cci.run performed prior.
    pval_adj_cutoff: float
        Cutoff for spot to be significant based on adjusted p-value.
    correct_axis: str
        Either 'spot', 'LR', or None; former corrects for number of LRs tested
        in each spot, middle corrects for number of spots tested per LR, and
        latter performs no adjustment (uses p-value for significant testing)
    adj_method: str
        Any method supported by statsmodels.stats.multitest.multipletests;
        https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html
    Returns
    -------
    adata: AnnData
        Adjusts all of the LR results; warning, does not adjust
        celltype-celltype results from running ran st.tl.run_cci downstream.
    """
    if "lr_summary" not in adata.uns:
        raise Exception("Need to run st.tl.cci.run first.")

    scores = adata.obsm["lr_scores"]
    sig_scores = scores.copy()
    ps = adata.obsm["p_vals"]
    padjs = np.ones(ps.shape)
    if correct_axis == "spot":
        for spot_i in range(ps.shape[0]):
            lr_indices = np.where(scores[spot_i, :] > 0)[0]
            if len(lr_indices) > 0:
                spot_ps = ps[spot_i, lr_indices]
                spot_padjs = multipletests(spot_ps, method=adj_method)[1]
                padjs[spot_i, lr_indices] = spot_padjs
                sig_scores[spot_i, lr_indices[spot_padjs >= pval_adj_cutoff]] = 0
    elif correct_axis == "LR":
        for lr_i in range(ps.shape[1]):
            spot_indices = np.where(scores[:, lr_i] > 0)[0]
            if len(spot_indices) > 0:
                lr_ps = ps[spot_indices, lr_i]
                spot_padjs = multipletests(lr_ps, method=adj_method)[1]
                padjs[spot_indices, lr_i] = spot_padjs
                sig_scores[spot_indices[spot_padjs >= pval_adj_cutoff], lr_i] = 0
    elif type(correct_axis) == type(None):
        padjs = ps.copy()
        sig_scores[padjs >= pval_adj_cutoff] = 0
    else:
        raise Exception(
            f"Invalid correct_axis input, must be one of: " f"'LR', 'spot', or None"
        )

    # Counting spots significant per lr #
    lr_counts = (padjs < pval_adj_cutoff).sum(axis=0)
    lr_counts_pval = (ps < pval_adj_cutoff).sum(axis=0)

    # Re-ranking LRs based on these counts & updating LR ordering #
    adata.uns["lr_summary"].loc[:, "n_spots_sig"] = lr_counts
    adata.uns["lr_summary"].loc[:, "n_spots_sig_pval"] = lr_counts_pval
    new_order = np.argsort(-adata.uns["lr_summary"].loc[:, "n_spots_sig"].values)
    adata.uns["lr_summary"] = adata.uns["lr_summary"].iloc[new_order, :]
    print(f"Updated adata.uns[lr_summary]")
    scores_ordered = scores[:, new_order]
    sig_scores_ordered = sig_scores[:, new_order]
    ps_ordered = ps[:, new_order]
    padjs_ordered = padjs[:, new_order]
    log10padjs = -np.log10(padjs_ordered)

    keys = ["lr_scores", "lr_sig_scores", "p_vals", "p_adjs", "-log10(p_adjs)"]
    values = [scores_ordered, sig_scores_ordered, ps_ordered, padjs_ordered, log10padjs]
    for i in range(len(keys)):
        adata.obsm[keys[i]] = values[i]
        print(f"Updated adata.obsm[{keys[i]}]")


def run_lr_go(
    adata: AnnData,
    r_path: str,
    n_top: int = 100,
    bg_genes: np.array = None,
    min_sig_spots: int = 1,
    species: str = "human",
    p_cutoff: float = 0.01,
    q_cutoff: float = 0.5,
    onts: str = "BP",
    verbose: bool = True,
):
    """Runs a basic GO analysis on the genes in the top ranked LR pairs.
        Only supported for human and mouse species.

    Parameters
    ----------
    adata: AnnData
        Must have had st.tl.cci_rank.run() called prior.
    r_path: str
        Path to R, must have clusterProfiler, org.Mm.eg.db, and org.Hs.eg.db
        installed.
    bg_genes: np.array
        Genes to be used as the background. If None, defaults to all genes in
        lr database: 'connectomeDB2020_put'.
    n_top: int
        The top number of LR pairs to use.
    min_sig_spots: int
        Minimum no. of significant spots pairs must have to be considered.
    species: str
        Species to perform the GO testing for.
    p_cutoff: float
        P-value & P-adj cutoff below which results will be returned.
    q_cutoff: float
        Q-value cutoff below which results will be returned.
    onts: str
        As per clusterProfiler; One of "BP", "MF", and "CC" subontologies, or "ALL" for all three.
    Returns
    -------
    adata: AnnData
        Relevant information stored in adata.uns['lr_go']
    """
    #### Making sure inputted correct species ####
    all_species = ["human", "mouse"]
    if species not in all_species:
        raise Exception(f"Got {species} for species, must be one of " f"{all_species}")

    #### Getting the genes from the top LR pairs ####
    if "lr_summary" not in adata.uns:
        raise Exception("Need to run st.tl.cci.run first.")
    lrs = adata.uns["lr_summary"].index.values.astype(str)
    n_sig = adata.uns["lr_summary"].loc[:, "n_spots_sig"].values.astype(int)
    top_lrs = lrs[n_sig > min_sig_spots][0:n_top]
    top_genes = np.unique([lr.split("_") for lr in top_lrs])

    ## Determining the background genes if not inputted ##
    if type(bg_genes) == type(None):
        all_lrs = load_lrs("connectomeDB2020_put")
        bg_genes = np.unique([lr_.split("_") for lr_ in all_lrs])

    #### Running the GO analysis ####
    go_results = run_GO(
        top_genes,
        bg_genes,
        species,
        r_path,
        p_cutoff=p_cutoff,
        q_cutoff=q_cutoff,
        onts=onts,
    )
    adata.uns["lr_go"] = go_results
    if verbose:
        print("GO results saved to adata.uns['lr_go']")


################################################################################
# Functions for calling Celltype-Celltype interactions #
################################################################################
def run_cci(
    adata: AnnData,
    use_label: str,
    spot_mixtures: bool = False,
    min_spots: int = 3,
    sig_spots: bool = True,
    cell_prop_cutoff: float = 0.2,
    p_cutoff=0.05,
    n_perms=100,
    verbose: bool = True,
):
    """Calls significant celltype-celltype interactions based on cell-type data randomisation.

    Parameters
    ----------
    adata: AnnData
        Must have had st.tl.cci_rank.run() called prior.
    use_label: str
        If !spot_mixtures, is a key in adata.obs, else key in adata.uns.
        Note if spot_mixtures specified, must have both the deconvolution data
        in adata.uns[use_label] and the dominant cell type per spot stored in
        adata.obs[use_label]. See tutorial for example.
    spot_mixtures: bool
        If true, indicates using deconvolution data, hence use_label
        refers to adata.uns.
    min_spots: int
        Specifies the minimum number of spots where LR score present to
        include in subsequent analysis.
    sig_spots: bool
        If true, only consider edges which include a signficant spot from
        calling st.tl.cci.run()
    cell_prop_cutoff: float
        Only relevant if spot_mixtures==True, indicates cutoff where cell type
        considered found in spot.
    p_cutoff: float
        Value at which p is considered significant.
    n_perms: int
        Number of randomisations of cell data to generate p-values.
    verbose: bool
        True if print dialogue to user during run-time.
    Returns
    -------
    adata: AnnData
        Relevant information stored
            adata.uns['lr_summary']
                Additional columns; f"n_cci_sig_{use_label}",
                f"n-spot_cci_{use_label}", f"n-spot_cci_sig_{use_label}".
                Former is the no. of CCIs significant for the LR, middle is
                the no. of individual spot-spot interactions across all CCIs for
                LR, and latter is the no. of significant individual spot
                interactions.
            adata.uns
                Dataframes added:
                    f"lr_cci_raw_{use_label}"
                        The raw count of spot-spot interactions across all LR
                        pairs for each possible CCI.
                    f"lr_cci_raw_{use_label}"
                        The count of significant spot-spot interactions across
                        all LR pairs for each possible CCI.
                Dictionaries added:
                    f"per_lr_cci_pvals_{use_label}"
                        Each key refers to a LR, with the value being a dataframe
                        listing the p-values for each potential CCI.
                    f"per_lr_cci_raw_{use_label}"
                        Each key refers to a LR, with the value being a dataframe
                        listing the count of spot-spot interactions via the LR in
                        significant LR neighbourhoods stratified by each
                        celltype-celltype combination.
                    f"per_lr_cci_{use_label}"
                        The same as f"per_lr_cci_raw_{use_label}", except
                        subsetted to significant CCIs.
    """
    ran_lr = "lr_summary" in adata.uns
    ran_sig = False if not ran_lr else "n_spots_sig" in adata.uns["lr_summary"].columns
    if not ran_lr and not ran_sig:
        raise Exception(
            "No LR results testing results found, " "please run st.tl.cci.run first"
        )

    # Ensuring compatibility with current way of adding label_transfer to object
    if use_label == "label_transfer" or use_label == "predictions":
        obs_key, uns_key = "predictions", "label_transfer"
    else:
        obs_key, uns_key = use_label, use_label

    # Getting the cell/tissue types that we are actually testing #
    if obs_key not in adata.obs:
        raise Exception(
            f"Missing {obs_key} from adata.obs, need this even if "
            f"using mixture mode."
        )
    tissue_types = adata.obs[obs_key].values.astype(str)
    all_set = np.unique(tissue_types)

    # Mixture mode
    mix_mode = spot_mixtures and uns_key in adata.uns
    if not mix_mode and spot_mixtures:
        print(
            f"Warning: specified spot_mixtures but no deconvolution data in "
            f"adata.uns['{uns_key}'].\nFalling back to discrete mode."
        )
    if mix_mode:  # Checking the deconvolution results stored correctly.
        cols_present = np.all(
            [cell_type in adata.uns[uns_key] for cell_type in all_set]
        )
        rows_present = np.all(adata.uns[uns_key].index.values == adata.obs_names.values)
        msg = f"Cell type scores misformatted in adata.uns[{uns_key}]:\n"
        if not cols_present or not rows_present:
            if not cols_present:
                msg = (
                    msg + f"Cell types missing from adata.uns[{uns_key}] columns:\n"
                    f"{[cell for cell in all_set if cell not in adata.uns[uns_key]]}\n"
                )
            elif not rows_present:
                msg = msg + "Rows do not correspond to adata.obs_names.\n"
            raise Exception(msg)

        #### Checking for case where have cell types that are never dominant
        #### in a spot, so need to include these in all_set
        if len(all_set) < adata.uns[uns_key].shape[1]:
            all_set = adata.uns[uns_key].columns.values.astype(str)

    # Getting minimum necessary information for edge counting #
    (
        spot_bcs,
        cell_data,
        neighbourhood_bcs,
        neighbourhood_indices,
    ) = get_data_for_counting(adata, use_label, mix_mode, all_set)

    lr_summary = adata.uns["lr_summary"]
    col_i = 1 if sig_spots else 0
    col = "lr_sig_scores" if sig_spots else "lr_scores"
    best_lr_indices = np.where(lr_summary.values[:, col_i] > min_spots)[0]
    best_lrs = lr_summary.index.values[best_lr_indices]
    lr_genes = np.unique([lr.split("_") for lr in best_lrs])
    if len(lr_genes) == 0:
        raise Exception(
            "No LR pairs returned with current filtering params; \n"
            "may need to adjust min_spots, sig_spots parameters, "
            "or re-run st.tl.cci.run with more relaxed parameters."
        )
    lr_expr = adata[:, lr_genes].to_df()

    # Sig-CCIs across all LRs #
    all_matrix = np.zeros((len(all_set), len(all_set)), dtype=int)
    # CCIs across all LRs #
    raw_matrix = np.zeros((len(all_set), len(all_set)), dtype=int)
    per_lr_cci = {}  # Per LR significant CCI counts #
    per_lr_cci_pvals = {}  # Per LR CCI p-values #
    per_lr_cci_raw = {}  # Per LR raw CCI counts #
    lr_n_spot_cci = np.zeros((lr_summary.shape[0]))
    lr_n_spot_cci_sig = np.zeros((lr_summary.shape[0]))
    lr_n_cci_sig = np.zeros((lr_summary.shape[0]))
    for i, best_lr in enumerate(best_lrs):
        l, r = best_lr.split("_")

        L_bool = lr_expr.loc[:, l].values > 0
        R_bool = lr_expr.loc[:, r].values > 0
        lr_index = np.where(adata.uns["lr_summary"].index.values == best_lr)[0][0]
        sig_bool = adata.obsm[col][:, lr_index] > 0

        int_matrix = get_interaction_matrix(
            cell_data,
            neighbourhood_bcs,
            neighbourhood_indices,
            all_set,
            sig_bool,
            L_bool,
            R_bool,
            cell_prop_cutoff,
        ).astype(int)

        int_pvals = get_interaction_pvals(
            int_matrix,
            n_perms,
            cell_data,
            neighbourhood_bcs,
            neighbourhood_indices,
            all_set,
            sig_bool,
            L_bool,
            R_bool,
            cell_prop_cutoff,
        )

        # Setting spot counts to 0 for non-significant ccis #
        sig_int_matrix = int_matrix.copy()
        sig_int_matrix[int_pvals > p_cutoff] = 0

        # Summarising n-spots sig/non-sig across ccis & total cci_rank for lr pair
        lr_n_spot_cci[best_lr_indices[i]] = int_matrix.sum()
        lr_n_spot_cci_sig[best_lr_indices[i]] = sig_int_matrix.sum()
        lr_n_cci_sig[best_lr_indices[i]] = (int_pvals < p_cutoff).sum()

        raw_matrix += int_matrix
        all_matrix += sig_int_matrix
        int_df = pd.DataFrame(int_matrix, index=all_set, columns=all_set)
        sig_int_df = pd.DataFrame(sig_int_matrix, index=all_set, columns=all_set)
        pval_df = pd.DataFrame(int_pvals, index=all_set, columns=all_set)
        per_lr_cci[best_lr] = sig_int_df
        per_lr_cci_pvals[best_lr] = pval_df
        per_lr_cci_raw[best_lr] = int_df

    # Saving results to anndata #
    adata.uns["lr_summary"][f"n_cci_sig_{use_label}"] = lr_n_cci_sig
    adata.uns["lr_summary"][f"n-spot_cci_{use_label}"] = lr_n_spot_cci
    adata.uns["lr_summary"][f"n-spot_cci_sig_{use_label}"] = lr_n_spot_cci_sig
    adata.uns[f"lr_cci_{use_label}"] = pd.DataFrame(
        all_matrix, index=all_set, columns=all_set
    )
    adata.uns[f"lr_cci_raw_{use_label}"] = pd.DataFrame(
        raw_matrix, index=all_set, columns=all_set
    )
    adata.uns[f"per_lr_cci_{use_label}"] = per_lr_cci
    adata.uns[f"per_lr_cci_pvals_{use_label}"] = per_lr_cci_pvals
    adata.uns[f"per_lr_cci_raw_{use_label}"] = per_lr_cci_raw
    if verbose:
        print(
            f"Significant counts of cci_rank interactions for all LR pairs in "
            f"{f'lr_cci_{use_label}'}"
        )
        print(
            f"Significant counts of cci_rank interactions for each LR pair "
            f"stored in dictionary {f'per_lr_cci_{use_label}'}"
        )



from scipy.stats import spearmanr
import numpy as np
import pandas as pd
import warnings
import networkx as nx
from ...utils import _read_graph

warnings.filterwarnings("ignore", category=RuntimeWarning)


def detect_transition_markers_clades(
    adata,
    clade,
    cutoff_spearman=0.4,
    cutoff_pvalue=0.05,
    screening_genes=None,
    use_raw_count=False,
):
    """\
    Transition markers detection of a clade.

    Parameters
    ----------
    adata
        Annotated data matrix.
    clade
        Name of a clade user wants to detect transition markers.
    cutoff_spearman
        The threshold of correlation coefficient.
    cutoff_pvalue
        The threshold of p-value.
    screening_genes
        List of customised genes.
    use_raw_count
        True if user wants to use raw layer data.
    Returns
    -------
    Anndata
    """

    print("Detecting the transition markers of clade_" + str(clade) + "...")

    tmp = _read_graph(adata, "PTS_graph")

    G = tmp.copy()

    remove = [edge for edge in G.edges if 9999 in edge]
    G.remove_edges_from(remove)
    G.remove_node(9999)

    nodes = []
    for edge in G.edges([clade]):
        nodes.append(edge[0])
        nodes.append(edge[1])
    nodes = list(set(nodes))

    assert clade in nodes, "Please choose the right clade!"

    query_adata = adata[adata.obs.query(create_query(nodes)).index]

    spearman_result = get_rank_cor(
        query_adata, screening_genes=screening_genes, use_raw_count=use_raw_count
    )

    spearman_result = spearman_result[spearman_result["p-value"] < cutoff_pvalue]
    positive = spearman_result[spearman_result["score"] >= cutoff_spearman].sort_values(
        "score", ascending=False
    )
    negative = spearman_result[
        spearman_result["score"] <= cutoff_spearman * (-1)
    ].sort_values("score")

    result = pd.concat([positive, negative])

    adata.uns["clade_" + str(clade)] = result

    print(
        "Transition markers result is stored in adata.uns['clade_" + str(clade) + "']"
    )


def detect_transition_markers_branches(
    adata,
    branch,
    cutoff_spearman=0.4,
    cutoff_pvalue=0.05,
    screening_genes=None,
    use_raw_count=False,
):
    """\
    Transition markers detection of a branch.

    Parameters
    ----------
    adata
        Annotated data matrix.
    branch
        Name of a branch user wants to detect transition markers.
    cutoff_spearman
        The threshold of correlation coefficient.
    cutoff_pvalue
        The threshold of p-value.
    screening_genes
        List of customised genes.
    use_raw_count
        True if user wants to use raw layer data.
    Returns
    -------
    Anndata
    """
    print(
        "Detecting the transition markers of branch_"
        + "_".join(np.array(branch).astype(str))
    )

    query_adata = adata[adata.obs.query(create_query(branch)).index]

    spearman_result = get_rank_cor(
        query_adata, screening_genes=screening_genes, use_raw_count=use_raw_count
    )

    spearman_result = spearman_result[spearman_result["p-value"] < cutoff_pvalue]
    positive = spearman_result[spearman_result["score"] >= cutoff_spearman].sort_values(
        "score", ascending=False
    )
    negative = spearman_result[
        spearman_result["score"] <= cutoff_spearman * (-1)
    ].sort_values("score")

    result = pd.concat([positive, negative])

    adata.uns["branch_" + "_".join(np.array(branch).astype(str))] = result

    print(
        "Transition markers result is stored in adata.uns['branch_"
        + "_".join(np.array(branch).astype(str))
        + "']"
    )


def create_query(list_sub_clusters):
    ini = ""
    for sub in list_sub_clusters:
        ini = ini + 'sub_cluster_labels == "' + str(sub) + '" | '
    return ini[:-2]


def get_rank_cor(adata, screening_genes=None, use_raw_count=True):
    if use_raw_count:
        tmp = adata.copy()
        tmp.X = adata.layers["raw_count"]
        tmp = tmp.to_df()
    else:
        tmp = adata.to_df()
    if screening_genes != None:
        tmp = tmp[screening_genes]
    dpt = adata.obs["dpt_pseudotime"].values
    genes = []
    score = []
    pvalue = []
    for gene in list(adata.var.index):
        genes.append(gene)
        score.append(spearmanr(tmp[gene].values, dpt)[0])
        pvalue.append(spearmanr(tmp[gene].values, dpt)[1])
    import pandas as pd

    final = pd.DataFrame({"gene": genes, "score": score, "p-value": pvalue})
    return final



""" Wrapper for performing the LR GO analysis.
"""

import os
import stlearn.tools.microenv.cci.r_helpers as rhs


def run_GO(genes, bg_genes, species, r_path, p_cutoff=0.01, q_cutoff=0.5, onts="BP"):
    """Running GO term analysis."""

    # Setting up the R environment #
    rhs.rpy2_setup(r_path)

    # Adding the source R code #
    r = rhs.ro.r
    path = os.path.dirname(os.path.realpath(__file__))
    r["source"](path + "/go.R")

    # Loading the GO analysis function #
    go_analyse_r = rhs.ro.globalenv["GO_analyse"]

    # Running the function on the genes #
    genes_r = rhs.ro.StrVector(genes)
    if type(bg_genes) != type(None):
        bg_genes_r = rhs.ro.StrVector(bg_genes)
    else:
        bg_genes_r = rhs.ro.r["as.null"]()
    p_cutoff_r = rhs.ro.FloatVector([p_cutoff])
    q_cutoff_r = rhs.ro.FloatVector([q_cutoff])
    onts_r = rhs.ro.StrVector([onts])
    go_results_r = go_analyse_r(
        genes_r, bg_genes_r, species, p_cutoff_r, q_cutoff_r, onts_r
    )
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        go_results = rhs.ro.conversion.rpy2py(go_results_r)

    return go_results



import matplotlib.pyplot as plt
from decimal import Decimal
from typing import Optional, Union
from anndata import AnnData


def DE_transition_plot(
    adata: AnnData,
    top_genes: int = 10,
    font_size: int = 6,
    name: str = None,
    dpi: int = 150,
    output: str = None,
) -> Optional[AnnData]:

    """\
    Differential expression between transition markers.

    Parameters
    ----------
    adata
        Annotated data matrix.
    top_genes
        Number of genes using to plot.
    font_size
        Size of the font.
    name
        Name of the output figure file.
    dpi
        DPI of the output figure.
    output
        Save the figure as file or not.
    Returns
    -------
    Figure object
    """

    trajectories = adata.uns["compare_result"]["trajectories"]
    pos_1 = (
        adata.uns[trajectories[0]]
        .set_index("gene")
        .loc[adata.uns["compare_result"]["pos_1"][:top_genes]]
        .iloc[::-1]
    )
    pos_2 = (
        adata.uns[trajectories[1]]
        .set_index("gene")
        .loc[adata.uns["compare_result"]["pos_2"][:top_genes]]
        .iloc[::-1]
    )
    neg_1 = (
        adata.uns[trajectories[0]]
        .set_index("gene")
        .loc[adata.uns["compare_result"]["neg_1"][:top_genes]]
        .iloc[::-1]
    )
    neg_2 = (
        adata.uns[trajectories[1]]
        .set_index("gene")
        .loc[adata.uns["compare_result"]["neg_2"][:top_genes]]
        .iloc[::-1]
    )

    y = range(top_genes)
    x1 = list(neg_1["score"])
    x2 = list(pos_1["score"])
    x3 = list(neg_2["score"])
    x4 = list(pos_2["score"])

    if len(x1) < top_genes:
        for i in range(len(x1), top_genes):
            x1.append(0)
    if len(x2) < top_genes:
        for i in range(len(x2), top_genes):
            x2.append(0)
    if len(x3) < top_genes:
        for i in range(len(x3), top_genes):
            x3.append(0)
    if len(x4) < top_genes:
        for i in range(len(x4), top_genes):
            x4.append(0)

    fig, axes = plt.subplots(ncols=2, nrows=2, sharey=True)
    fig.subplots_adjust(wspace=0, hspace=0.8)
    axes[0][0].barh(y, x1, align="center", color="#fb687a")
    axes[0][1].barh(y, x2, align="center", color="#31a2fb")
    axes[1][0].barh(y, x3, align="center", color="#fb687a")
    axes[1][1].barh(y, x4, align="center", color="#31a2fb")

    for i, x in enumerate([x1, x2, x3, x4]):
        if all(value == 0 for value in x):
            if i == 0:
                axes[0][0].get_xaxis().set_ticks([])
            if i == 1:
                axes[0][1].get_xaxis().set_ticks([])
            if i == 2:
                axes[1][0].get_xaxis().set_ticks([])
            if i == 3:
                axes[1][1].get_xaxis().set_ticks([])

    axes[0][0].spines["left"].set_visible(False)
    axes[0][0].spines["right"].set_visible(False)
    axes[0][0].spines["top"].set_visible(False)
    axes[1][0].spines["left"].set_visible(False)
    axes[1][0].spines["right"].set_visible(False)
    axes[1][0].spines["top"].set_visible(False)

    axes[0][1].spines["right"].set_visible(False)
    axes[0][1].spines["top"].set_visible(False)
    axes[0][1].spines["left"].set_visible(False)
    axes[1][1].spines["right"].set_visible(False)
    axes[1][1].spines["top"].set_visible(False)
    axes[1][1].spines["left"].set_visible(False)

    axes[0][0].get_yaxis().set_ticks([])
    axes[1][0].get_yaxis().set_ticks([])
    axes[0][0].tick_params(axis="both", which="both", length=0)
    axes[1][0].tick_params(axis="both", which="both", length=0)

    axes[0][1].get_yaxis().set_ticks([])
    axes[1][1].get_yaxis().set_ticks([])
    axes[0][1].tick_params(axis="both", which="both", length=0)
    axes[1][1].tick_params(axis="both", which="both", length=0)

    rects = axes[0][1].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = pos_1.index[i]
            p_value = "{:.2E}".format(Decimal(str(pos_1["p-value"][i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "left", "verticalalignment": "center"}
        axes[0][1].text(
            rect.get_x() + rect.get_width() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=font_size
        )
        axes[0][1].text(
            rect.get_x() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=font_size
        )

    rects = axes[0][0].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = neg_1.index[i]
            p_value = "{:.2E}".format(Decimal(str(neg_1["p-value"][i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "right", "verticalalignment": "center"}
        axes[0][0].text(
            rect.get_x() + rect.get_width() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=font_size
        )
        axes[0][0].text(
            rect.get_x() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=font_size
        )

    rects = axes[1][1].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = pos_2.index[i]
            p_value = "{:.2E}".format(Decimal(str(pos_2["p-value"][i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "left", "verticalalignment": "center"}
        axes[1][1].text(
            rect.get_x() + rect.get_width() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=font_size
        )
        axes[1][1].text(
            rect.get_x() + 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=font_size
        )

    rects = axes[1][0].patches
    for i, rect in enumerate(rects):
        try:
            gene_name = neg_2.index[i]
            p_value = "{:.2E}".format(Decimal(str(neg_2["p-value"][i])))
        except:
            gene_name = ""
            p_value = ""
        alignment = {"horizontalalignment": "right", "verticalalignment": "center"}
        axes[1][0].text(
            rect.get_x() + rect.get_width() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            gene_name,
            **alignment,
            size=font_size
        )
        axes[1][0].text(
            rect.get_x() - 0.01,
            rect.get_y() + rect.get_height() / 2.0,
            p_value,
            color="w",
            **alignment,
            size=font_size
        )

    plt.figtext(
        0.5,
        0.5,
        "Markers of " + trajectories[0] + " compared to " + trajectories[1],
        ha="center",
        va="center",
    )
    plt.figtext(
        0.5,
        0.0,
        "Markers of " + trajectories[1] + " compared to " + trajectories[0],
        ha="center",
        va="center",
    )
    plt.show()
    if output is not None:
        if name is not None:
            plt.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)



import scanpy as sc
from .._settings import settings
from pathlib import Path
from anndata import AnnData


def example_bcba() -> AnnData:
    """\
    Download processed BCBA data (10X genomics published data).
    Reference: https://support.10xgenomics.com/spatial-gene-expression/datasets/1.1.0/V1_Breast_Cancer_Block_A_Section_1
    """
    settings.datasetdir.mkdir(exist_ok=True)
    filename = settings.datasetdir / "example_bcba.h5"
    url = "https://www.dropbox.com/s/u3m2f16mvdom1am/example_bcba.h5ad?dl=1"
    if not filename.is_file():
        sc.readwrite._download(url=url, path=filename)
    adata = sc.read_h5ad(filename)
    return adata



from __future__ import division
import numpy as np
import pandas as pd
from PIL import Image
from stlearn.tools.microenv.cci.het import get_edges

from bokeh.plotting import (
    figure,
    show,
    ColumnDataSource,
    curdoc,
)
from bokeh.models import (
    BoxSelectTool,
    LassoSelectTool,
    CustomJS,
    Div,
    Paragraph,
    LinearColorMapper,
    Slider,
    Select,
    AutocompleteInput,
    ColorBar,
    TextInput,
    BasicTicker,
    HoverTool,
    ZoomOutTool,
    CheckboxGroup,
    Arrow,
    VeeHead,
    Button,
    Dropdown,
    Div,
)

from bokeh.models.widgets import DataTable, DateFormatter, TableColumn
from anndata import AnnData
from bokeh.palettes import (
    Spectral11,
    Viridis256,
    Reds256,
    Blues256,
    Magma256,
    Category20,
)
from bokeh.layouts import column, row, grid
from collections import OrderedDict
from bokeh.application import Application
from bokeh.application.handlers import FunctionHandler
from stlearn.classes import Spatial
from typing import Optional
from stlearn.utils import _read_graph
import scanpy as sc


class BokehGenePlot(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
    ):
        super().__init__(
            adata,
        )
        # Open image, and make sure it's RGB*A*
        image = (self.img * 255).astype(np.uint8)

        img_pillow = Image.fromarray(image).convert("RGBA")

        self.xdim, self.ydim = img_pillow.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        self.image = np.empty((self.ydim, self.xdim), dtype=np.uint32)
        view = self.image.view(dtype=np.uint8).reshape((self.ydim, self.xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view[:, :, :] = np.flipud(np.asarray(img_pillow))

        # Display the 32-bit RGBA image
        self.dim = max(self.xdim, self.ydim)
        gene_list = list(adata.var_names)

        self.data_alpha = Slider(
            title="Spot alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.tissue_alpha = Slider(
            title="Tissue alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.spot_size = Slider(
            title="Spot size", value=5.0, start=0, end=5.0, step=1.0
        )

        self.gene_select = AutocompleteInput(
            title="Gene:", value=gene_list[0], completions=gene_list, min_characters=1
        )

        color_list = ["Spectral", "Viridis", "Reds", "Blues", "Magma"]
        self.cmap_select = Select(
            title="Select color map:", value=color_list[0], options=color_list
        )

        self.output_backend = Select(
            title="Select output backend:", value="webgl", options=["webgl", "svg"]
        )

        self.menu = []

        for col in adata.obs.columns:
            if adata.obs[col].dtype.name == "category":
                if col != "sub_cluster_labels":
                    self.menu.append(col)
        if len(self.menu) != 0:
            self.use_label = Select(
                title="Select use_label:", value=self.menu[0], options=self.menu
            )
            inputs = column(
                self.gene_select,
                self.data_alpha,
                self.tissue_alpha,
                self.spot_size,
                self.cmap_select,
                self.use_label,
                self.output_backend,
            )
            self.layout = column(row(inputs, self.make_fig()), self.add_violin())
        else:
            inputs = column(
                self.gene_select,
                self.data_alpha,
                self.tissue_alpha,
                self.spot_size,
                self.cmap_select,
                self.output_backend,
            )
            self.layout = row(inputs, self.make_fig())

        # Make a tab with the layout
        # self.tab = Tabs(tabs = [Panel(child=self.layout, title="Gene plot")])

        def modify_fig(doc):

            doc.add_root(row(self.layout, width=800))

            self.data_alpha.on_change("value", self.update_data)
            self.tissue_alpha.on_change("value", self.update_data)
            self.spot_size.on_change("value", self.update_data)
            self.gene_select.on_change("value", self.update_data)
            self.cmap_select.on_change("value", self.update_data)
            self.output_backend.on_change("value", self.update_data)
            if len(self.menu) != 0:
                self.use_label.on_change("value", self.update_data)

        handler = FunctionHandler(modify_fig)

        self.app = Application(handler)

    def make_fig(self):

        fig = figure(
            title=self.gene_select.value,
            x_range=(0, self.dim - 150),
            y_range=(self.dim, 0),
            output_backend=self.output_backend.value,
            name="GenePlot",
            active_scroll="wheel_zoom",
        )

        colors = self._get_gene_expression([self.gene_select.value])

        s1 = ColumnDataSource(data=dict(x=self.imagecol, y=self.imagerow, color=colors))

        fig.image_rgba(
            image=[self.image],
            x=0,
            y=self.xdim,
            dw=self.ydim,
            dh=self.xdim,
            global_alpha=self.tissue_alpha.value,
        )
        if self.cmap_select.value == "Spectral":
            cmap = Spectral11
        elif self.cmap_select.value == "Viridis":
            cmap = Viridis256
        elif self.cmap_select.value == "Reds":
            cmap = list(Reds256)
            cmap.reverse()
        elif self.cmap_select.value == "Blues":
            cmap = list(Blues256)
            cmap.reverse()
        elif self.cmap_select.value == "Magma":
            cmap = Magma256

        color_mapper = LinearColorMapper(
            palette=cmap, low=min(s1.data["color"]), high=max(s1.data["color"])
        )

        fig.circle(
            "x",
            "y",
            color={"field": "color", "transform": color_mapper},
            size=self.spot_size.value,
            source=s1,
            fill_alpha=self.data_alpha.value,
            line_alpha=self.data_alpha.value,
        )

        color_bar = ColorBar(
            color_mapper=color_mapper, ticker=BasicTicker(), location=(-20, 0)
        )
        fig.add_layout(color_bar, "right")

        fig.toolbar.logo = None
        fig.xaxis.visible = False
        fig.yaxis.visible = False
        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.outline_line_alpha = 0
        fig.add_tools(LassoSelectTool())
        fig.add_tools(BoxSelectTool())
        fig.add_tools(HoverTool())

        hover = fig.select(dict(type=HoverTool))
        hover.tooltips = OrderedDict(
            [
                ("Spot", "$index"),
                ("X location", "@x{1.11}"),
                ("Y location", "@y{1.11}"),
                ("Gene expression", "@color{1.11}"),
            ]
        )

        return fig

    def add_violin(self):
        violin = self.create_violin(
            self.adata[0],
            gene_symbol=self.gene_select.value,
            use_label=self.use_label.value,
        )

        violin = (np.array(violin)).astype(np.uint8)

        img_pillow2 = Image.fromarray(violin).convert("RGBA")

        xdim, ydim = img_pillow2.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        image2 = np.empty((ydim, xdim), dtype=np.uint32)
        view2 = image2.view(dtype=np.uint8).reshape((ydim, xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view2[:, :, :] = np.flipud(np.asarray(img_pillow2))

        p = figure(
            plot_width=910,
            plot_height=int(910 / xdim * ydim) + 5,
            output_backend=self.output_backend.value,
        )

        # must give a vector of images
        p.image_rgba(image=[image2], x=0, y=1, dw=xdim, dh=ydim)

        p.toolbar.logo = None
        p.xaxis.visible = False
        p.yaxis.visible = False
        p.xgrid.grid_line_color = None
        p.ygrid.grid_line_color = None
        p.outline_line_alpha = 0

        return p

    def update_data(self, attrname, old, new):

        if len(self.menu) != 0:
            self.layout.children[0].children[1] = self.make_fig()
            self.layout.children[1] = self.add_violin()
        else:
            self.layout.children[1] = self.make_fig()

    def _get_gene_expression(self, gene_symbols):

        if gene_symbols[0] not in self.adata[0].var_names:
            raise ValueError(
                gene_symbols[0] + " is not exist in the data, please try another gene"
            )

        colors = self.adata[0][:, gene_symbols].to_df().iloc[:, -1]

        return colors

    def fig2img(self, fig):
        """Convert a Matplotlib figure to a PIL Image and return it"""
        import io

        buf = io.BytesIO()
        fig.savefig(buf, bbox_inches="tight", pad_inches=0, dpi=120)
        buf.seek(0)
        img = Image.open(buf)
        return img

    def create_violin(self, adata, gene_symbol, use_label):
        import matplotlib.pyplot as plt

        plt.rc("font", size=5)

        fig, ax = plt.subplots(figsize=(8, 5))
        sc.pl.violin(
            adata, keys=gene_symbol, groupby=use_label, rotation=45, ax=ax, show=False
        )

        # Hide the right and top spines
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        # Only show ticks on the left and bottom spines
        ax.yaxis.set_ticks_position("left")
        ax.xaxis.set_ticks_position("bottom")

        plt.close(fig)

        violin = self.fig2img(fig)

        return violin


class BokehClusterPlot(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
    ):
        super().__init__(adata)

        # Open image, and make sure it's RGB*A*
        image = (self.img * 255).astype(np.uint8)

        img_pillow = Image.fromarray(image).convert("RGBA")

        self.xdim, self.ydim = img_pillow.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        self.image = np.empty((self.ydim, self.xdim), dtype=np.uint32)
        view = self.image.view(dtype=np.uint8).reshape((self.ydim, self.xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view[:, :, :] = np.flipud(np.asarray(img_pillow))

        # Display the 32-bit RGBA image
        self.dim = max(self.xdim, self.ydim)

        menu = []

        for col in adata.obs.columns:
            if adata.obs[col].dtype.name == "category":
                if col != "sub_cluster_labels":
                    menu.append(col)

        self.use_label = Select(title="Select use_label:", value=menu[0], options=menu)

        # Initialize the color
        from stlearn.plotting.cluster_plot import cluster_plot

        if len(adata.obs[self.use_label.value].cat.categories) <= 20:
            cluster_plot(adata, use_label=self.use_label.value, show_plot=False)
        else:
            cluster_plot(
                adata,
                use_label=self.use_label.value,
                show_plot=False,
                cmap="default_102",
            )

        self.data_alpha = Slider(
            title="Spot alpha", value=1.0, start=0, end=1.0, step=0.1
        )

        self.tissue_alpha = Slider(
            title="Tissue alpha", value=1.0, start=0, end=1.0, step=0.1
        )

        self.spot_size = Slider(
            title="Spot size", value=5.0, start=0, end=5.0, step=1.0
        )

        self.checkbox_group = CheckboxGroup(
            labels=["Show spatial trajectories"], active=[]
        )

        self.p = Paragraph(text="""Choose clusters:""", width=400, height=20)

        self.list_cluster = CheckboxGroup(
            labels=list(self.adata[0].obs[self.use_label.value].cat.categories),
            active=list(
                np.array(
                    range(0, len(self.adata[0].obs[self.use_label.value].unique()))
                )
            ),
        )

        self.n_top_genes = Slider(
            title="Number of top genes", value=5, start=1, end=10, step=1
        )

        color_list = ["bwr", "RdBu_r", "viridis", "Reds", "Blues", "magma"]
        self.cmap_select = Select(
            title="Select color map:", value=color_list[0], options=color_list
        )

        plot_types = ["dotplot", "stacked_violin", "matrixplot"]
        self.plot_select = Select(
            title="Select DEA plot type:", value=plot_types[0], options=plot_types
        )

        self.output_backend = Select(
            title="Select output backend:", value="webgl", options=["webgl", "svg"]
        )

        self.min_logfoldchange = TextInput(value="3", title="Min log fold-change")

        if "PTS_graph" in adata.uns:
            if "rank_genes_groups" in self.adata[0].uns:
                self.inputs = column(
                    self.use_label,
                    self.data_alpha,
                    self.tissue_alpha,
                    self.spot_size,
                    self.p,
                    self.list_cluster,
                    self.checkbox_group,
                    self.n_top_genes,
                    self.min_logfoldchange,
                    self.cmap_select,
                    self.plot_select,
                    self.output_backend,
                )
            else:
                self.inputs = column(
                    self.use_label,
                    self.data_alpha,
                    self.tissue_alpha,
                    self.spot_size,
                    self.p,
                    self.list_cluster,
                    self.checkbox_group,
                    self.output_backend,
                )

        else:
            if "rank_genes_groups" in self.adata[0].uns:
                self.inputs = column(
                    self.use_label,
                    self.data_alpha,
                    self.tissue_alpha,
                    self.spot_size,
                    self.p,
                    self.list_cluster,
                    self.n_top_genes,
                    self.min_logfoldchange,
                    self.cmap_select,
                    self.plot_select,
                    self.output_backend,
                )
            else:
                self.inputs = column(
                    self.use_label,
                    self.data_alpha,
                    self.tissue_alpha,
                    self.spot_size,
                    self.p,
                    self.list_cluster,
                    self.output_backend,
                )

        if "rank_genes_groups" in self.adata[0].uns:
            if (
                self.use_label.value
                == self.adata[0].uns["rank_genes_groups"]["params"]["groupby"]
            ):
                self.layout = column(row(self.inputs, self.make_fig()), self.add_dea())
            else:
                self.layout = column(row(self.inputs, self.make_fig()), Div(text=""))
        else:
            self.layout = row(self.inputs, self.make_fig())

        def modify_fig(doc):
            doc.add_root(row(self.layout, width=800))
            self.use_label.on_change("value", self.update_list)
            self.use_label.on_change("value", self.update_data)

            self.data_alpha.on_change("value", self.update_data)
            self.tissue_alpha.on_change("value", self.update_data)
            self.spot_size.on_change("value", self.update_data)
            self.list_cluster.on_change("active", self.update_data)
            if "PTS_graph" in self.adata[0].uns:
                self.checkbox_group.on_change("active", self.update_data)

            self.output_backend.on_change("value", self.update_data)
            if "rank_genes_groups" in self.adata[0].uns:
                self.n_top_genes.on_change("value", self.update_data)
                self.cmap_select.on_change("value", self.update_data)
                self.plot_select.on_change("value", self.update_data)
                self.min_logfoldchange.on_change("value", self.update_data)

        handler = FunctionHandler(modify_fig)
        self.app = Application(handler)

    def update_list(self, attrname, old, name):

        # Initialize the color
        from stlearn.plotting.cluster_plot import cluster_plot

        cluster_plot(self.adata[0], use_label=self.use_label.value, show_plot=False)

        # self.list_cluster = CheckboxGroup(
        #     labels=list(self.adata[0].obs[self.use_label.value].cat.categories),
        #     active=list(
        #         np.array(range(0, len(self.adata[0].obs[self.use_label.value].unique())))
        #     ),
        # )
        self.list_cluster.labels = list(
            self.adata[0].obs[self.use_label.value].cat.categories
        )
        self.list_cluster.active = list(
            np.array(range(0, len(self.adata[0].obs[self.use_label.value].unique())))
        )

    def update_data(self, attrname, old, new):

        if "rank_genes_groups" in self.adata[0].uns:
            if (
                self.use_label.value
                == self.adata[0].uns["rank_genes_groups"]["params"]["groupby"]
            ):
                self.layout.children[0].children[1] = self.make_fig()
                self.layout.children[1] = self.add_dea()
            else:
                if len(self.layout.children) > 1:
                    self.layout.children[1] = Div(text="")
                    self.layout.children[0].children[1] = self.make_fig()
        else:
            self.layout.children[1] = self.make_fig()

    def make_fig(self):
        fig = figure(
            title="Cluster plot",
            x_range=(0, self.dim - 150),
            y_range=(self.dim, 0),
            output_backend=self.output_backend.value
            # Specifying xdim/ydim isn't quire right :-(
            # width=xdim, height=ydim,
        )

        fig.image_rgba(
            image=[self.image],
            x=0,
            y=self.xdim,
            dw=self.ydim,
            dh=self.xdim,
            global_alpha=self.tissue_alpha.value,
        )

        # Get query clusters
        command = []
        for i in self.list_cluster.active:
            command.append(
                self.use_label.value
                + ' == "'
                + self.adata[0].obs[self.use_label.value].cat.categories[i]
                + '"'
            )
        tmp = self.adata[0].obs.query(" or ".join(command))

        tmp_adata = self.adata[0][tmp.index, :]

        x = tmp_adata.obsm["spatial"][:, 0] * self.scale_factor
        y = tmp_adata.obsm["spatial"][:, 1] * self.scale_factor

        category_items = self.adata[0].obs[self.use_label.value].cat.categories
        palette = self.adata[0].uns[self.use_label.value + "_colors"]
        colormap = dict(zip(category_items, palette))
        color = list(tmp[self.use_label.value].map(colormap))
        cluster = list(tmp[self.use_label.value])

        s1 = ColumnDataSource(data=dict(x=x, y=y, color=color, cluster=cluster))
        if len(category_items[0]) > 5:
            fig.scatter(
                x="x",
                y="y",
                source=s1,
                size=self.spot_size.value,
                color="color",
                # legend_group="cluster",
                fill_alpha=self.data_alpha.value,
                line_alpha=self.data_alpha.value,
            )
        else:
            fig.scatter(
                x="x",
                y="y",
                source=s1,
                size=self.spot_size.value,
                color="color",
                legend_group="cluster",
                fill_alpha=self.data_alpha.value,
                line_alpha=self.data_alpha.value,
            )

        if 0 in self.checkbox_group.active:
            tmp = _read_graph(self.adata[0], "PTS_graph")
            G = tmp.copy()

            remove = [edge for edge in G.edges if 9999 in edge]
            G.remove_edges_from(remove)
            G.remove_node(9999)
            centroid_dict = self.adata[0].uns["centroid_dict"]
            centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

            set_x = []
            set_y = []
            for edges in G.edges:
                fig.add_layout(
                    Arrow(
                        end=VeeHead(fill_color="#f4efd3", line_alpha=0, line_width=5),
                        x_start=centroid_dict[edges[0]][0],
                        y_start=centroid_dict[edges[0]][1],
                        x_end=centroid_dict[edges[1]][0],
                        y_end=centroid_dict[edges[1]][1],
                        line_color="#f4efd3",
                        line_width=5,
                    )
                )
                set_x.append(centroid_dict[edges[0]][0])
                set_y.append(centroid_dict[edges[0]][1])

            fig.circle(
                x=set_x,
                y=set_y,
                radius=50,
                color="#f4efd3",
                fill_alpha=0.3,
                line_alpha=0,
            )

        fig.toolbar.logo = None
        fig.xaxis.visible = False
        fig.yaxis.visible = False
        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.outline_line_alpha = 0
        fig.add_tools(LassoSelectTool())
        fig.add_tools(ZoomOutTool())
        fig.add_tools(HoverTool())
        fig.add_tools(BoxSelectTool())

        hover = fig.select(dict(type=HoverTool))
        hover.tooltips = OrderedDict(
            [
                ("Spot", "$index"),
                ("X location", "@x{1.11}"),
                ("Y location", "@y{1.11}"),
                ("Cluster", "@cluster"),
            ]
        )

        return fig

    def add_dea(self):
        dea = self.create_dea(self.adata[0])

        dea = (np.array(dea)).astype(np.uint8)

        img_pillow2 = Image.fromarray(dea).convert("RGBA")

        xdim, ydim = img_pillow2.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        image2 = np.empty((ydim, xdim), dtype=np.uint32)
        view2 = image2.view(dtype=np.uint8).reshape((ydim, xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view2[:, :, :] = np.flipud(np.asarray(img_pillow2))

        p = figure(
            plot_width=910,
            plot_height=int(910 / xdim * ydim) + 5,
            output_backend=self.output_backend.value,
        )

        # must give a vector of images
        p.image_rgba(image=[image2], x=0, y=1, dw=xdim, dh=ydim)

        p.toolbar.logo = None
        p.xaxis.visible = False
        p.yaxis.visible = False
        p.xgrid.grid_line_color = None
        p.ygrid.grid_line_color = None
        p.outline_line_alpha = 0

        return p

    def fig2img(self, fig):
        """Convert a Matplotlib figure to a PIL Image and return it"""
        import io

        buf = io.BytesIO()
        fig.savefig(buf, bbox_inches="tight", pad_inches=0, dpi=150)
        buf.seek(0)
        img = Image.open(buf)
        return img

    def create_dea(self, adata):
        import matplotlib.pyplot as plt

        plt.rc("font", size=12)

        fig, ax = plt.subplots(figsize=(12, 5))
        sc.tl.dendrogram(adata, groupby=self.use_label.value)
        if self.plot_select.value == "matrixplot":
            sc.pl.rank_genes_groups_matrixplot(
                adata,
                n_genes=self.n_top_genes.value,
                cmap=self.cmap_select.value,
                show=False,
                ax=ax,
                standard_scale="var",
                min_logfoldchange=float(self.min_logfoldchange.value),
                groups=self.adata[0]
                .obs[self.use_label.value]
                .cat.categories[self.list_cluster.active],
            )
        elif self.plot_select.value == "stacked_violin":
            sc.pl.rank_genes_groups_stacked_violin(
                adata,
                n_genes=self.n_top_genes.value,
                cmap=self.cmap_select.value,
                show=False,
                ax=ax,
                standard_scale="var",
                min_logfoldchange=float(self.min_logfoldchange.value),
                groups=self.adata[0]
                .obs[self.use_label.value]
                .cat.categories[self.list_cluster.active],
            )
        else:
            sc.pl.rank_genes_groups_dotplot(
                adata,
                n_genes=self.n_top_genes.value,
                cmap=self.cmap_select.value,
                show=False,
                ax=ax,
                standard_scale="var",
                min_logfoldchange=float(self.min_logfoldchange.value),
                groups=self.adata[0]
                .obs[self.use_label.value]
                .cat.categories[self.list_cluster.active],
            )

        # Hide the right and top spines
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        # Only show ticks on the left and bottom spines
        ax.yaxis.set_ticks_position("left")
        ax.xaxis.set_ticks_position("bottom")

        plt.close(fig)

        dea = self.fig2img(fig)

        return dea


class BokehLRPlot(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
    ):
        super().__init__(
            adata,
        )
        # Open image, and make sure it's RGB*A*
        image = (self.img * 255).astype(np.uint8)

        img_pillow = Image.fromarray(image).convert("RGBA")

        self.xdim, self.ydim = img_pillow.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        self.image = np.empty((self.ydim, self.xdim), dtype=np.uint32)
        view = self.image.view(dtype=np.uint8).reshape((self.ydim, self.xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view[:, :, :] = np.flipud(np.asarray(img_pillow))

        # Display the 32-bit RGBA image
        self.dim = max(self.xdim, self.ydim)
        # available_het = [] #Checking available LR scores
        # for key in list(self.adata[0].obsm.keys()):
        #     if len(self.adata[0].obsm[key].shape) == 1:
        #         available_het.append(key)
        lrs = list(adata.uns["lr_summary"].index.values.astype(str))

        self.data_alpha = Slider(
            title="Spot alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.tissue_alpha = Slider(
            title="Tissue alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.spot_size = Slider(
            title="Spot size", value=5.0, start=0, end=5.0, step=1.0
        )

        # self.het_select = AutocompleteInput(
        #     title="Het:",
        #     value=available_het[0],
        #     completions=available_het,
        #     min_characters=1,
        # )
        # self.lr_select = AutocompleteInput(
        #     title="Ligand-receptor:",
        #     value=lrs[0],
        #     completions=lrs,
        #     min_characters=1,
        # )
        self.lr_select = Select(
            title="Ligand-receptor:",
            value=lrs[0],
            options=lrs,
        )

        self.output_backend = Select(
            title="Select output backend:", value="webgl", options=["webgl", "svg"]
        )

        inputs = column(
            # self.het_select,
            self.lr_select,
            self.data_alpha,
            self.tissue_alpha,
            self.spot_size,
            self.output_backend,
        )

        self.layout = row(inputs, self.make_fig())

        # Make a tab with the layout
        # self.tab = Tabs(tabs = [Panel(child=self.layout, title="Gene plot")])

        def modify_fig(doc):

            doc.add_root(row(self.layout, width=800))

            self.data_alpha.on_change("value", self.update_data)
            self.tissue_alpha.on_change("value", self.update_data)
            self.spot_size.on_change("value", self.update_data)
            # self.het_select.on_change("value", self.update_data)
            self.lr_select.on_change("value", self.update_data)

        handler = FunctionHandler(modify_fig)

        self.app = Application(handler)

    def make_fig(self):

        fig = figure(
            title=self.lr_select.value,  # self.het_select.value,
            x_range=(0, self.dim - 150),
            y_range=(self.dim, 0),
            output_backend=self.output_backend.value,
        )

        colors = self._get_lr(self.lr_select.value)
        # colors = self._get_het(self.het_select.value)

        s1 = ColumnDataSource(data=dict(x=self.imagecol, y=self.imagerow, color=colors))

        fig.image_rgba(
            image=[self.image],
            x=0,
            y=self.xdim,
            dw=self.ydim,
            dh=self.xdim,
            global_alpha=self.tissue_alpha.value,
        )

        color_mapper = LinearColorMapper(
            palette=Spectral11, low=min(s1.data["color"]), high=max(s1.data["color"])
        )

        fig.circle(
            "x",
            "y",
            color={"field": "color", "transform": color_mapper},
            size=self.spot_size.value,
            source=s1,
            fill_alpha=self.data_alpha.value,
            line_alpha=self.data_alpha.value,
        )

        color_bar = ColorBar(
            color_mapper=color_mapper, ticker=BasicTicker(), location=(10, 0)
        )
        fig.add_layout(color_bar, "right")

        fig.toolbar.logo = None
        fig.xaxis.visible = False
        fig.yaxis.visible = False
        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.outline_line_alpha = 0
        fig.add_tools(LassoSelectTool())
        fig.add_tools(BoxSelectTool())
        fig.add_tools(HoverTool())

        hover = fig.select(dict(type=HoverTool))
        hover.tooltips = OrderedDict(
            [
                ("Spot", "$index"),
                ("X location", "@x{1.11}"),
                ("Y location", "@y{1.11}"),
                ("Values", "@color{1.11}"),
            ]
        )

        return fig

    def update_data(self, attrname, old, new):
        self.layout.children[1] = self.make_fig()

    def _get_het(self, het):

        if het not in self.adata[0].obsm:
            raise ValueError(het + " is not exist in the data, please try another het")

        colors = self.adata[0].obsm[het]

        return colors

    def _get_lr(self, lr):
        if lr not in self.adata[0].uns["lr_summary"].index:
            raise ValueError(lr + " is not exist in the data, please try another het")

        lr_bool = self.adata[0].uns["lr_summary"].index.values.astype(str) == lr
        lr_index = np.where(lr_bool)[0][0]
        colors = self.adata[0].obsm["lr_sig_scores"][:, lr_index]

        return colors


class BokehSpatialCciPlot(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
    ):
        super().__init__(
            adata,
        )
        # Open image, and make sure it's RGB*A*
        image = (self.img * 255).astype(np.uint8)

        img_pillow = Image.fromarray(image).convert("RGBA")

        self.xdim, self.ydim = img_pillow.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        self.image = np.empty((self.ydim, self.xdim), dtype=np.uint32)
        view = self.image.view(dtype=np.uint8).reshape((self.ydim, self.xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view[:, :, :] = np.flipud(np.asarray(img_pillow))

        # Display the 32-bit RGBA image
        self.dim = max(self.xdim, self.ydim)
        # available_het = [] #Checking available LR scores
        # for key in list(self.adata[0].obsm.keys()):
        #     if len(self.adata[0].obsm[key].shape) == 1:
        #         available_het.append(key)
        lrs = list(adata.uns["lr_summary"].index.values.astype(str))

        self.data_alpha = Slider(
            title="Spot alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.tissue_alpha = Slider(
            title="Tissue alpha", value=1.0, start=0, end=1.0, step=0.1
        )
        self.spot_size = Slider(
            title="Spot size", value=5.0, start=0, end=10.0, step=1.0
        )
        self.arrow_size = Slider(
            title="Arrow size", value=1.0, start=0, end=10.0, step=0.5
        )

        # Getting the annnotations for which CCI has been performed.. #
        annots = [
            opt.replace("lr_cci_", "")
            for opt in adata.uns.keys()
            if opt.startswith("lr_cci_") and "raw" not in opt
        ]
        self.annot_select = Select(
            title="Cell-type annotation select:",
            value=annots[0],
            options=annots,
        )
        self.lr_select = Select(
            title="Ligand-receptor:",
            value=lrs[0],
            options=lrs,
        )

        self.list_cluster = CheckboxGroup(
            labels=list(self.adata[0].obs[self.annot_select.value].cat.categories),
            active=list(
                np.array(
                    range(0, len(self.adata[0].obs[self.annot_select.value].unique()))
                )
            ),
        )

        self.output_backend = Select(
            title="Select output backend:", value="webgl", options=["webgl", "svg"]
        )

        inputs = column(
            # self.het_select,
            self.annot_select,
            self.lr_select,
            self.data_alpha,
            self.tissue_alpha,
            self.spot_size,
            self.arrow_size,
            self.output_backend,
        )

        self.layout = row(inputs, self.make_fig())

        # Make a tab with the layout
        # self.tab = Tabs(tabs = [Panel(child=self.layout, title="Gene plot")])

        def modify_fig(doc):

            doc.add_root(row(self.layout, width=800))

            self.data_alpha.on_change("value", self.update_data)
            self.tissue_alpha.on_change("value", self.update_data)
            self.spot_size.on_change("value", self.update_data)
            self.arrow_size.on_change("value", self.update_data)
            # self.het_select.on_change("value", self.update_data)
            self.annot_select.on_change("value", self.update_data)
            self.lr_select.on_change("value", self.update_data)

        handler = FunctionHandler(modify_fig)

        self.app = Application(handler)

    def make_fig(self):

        fig = figure(
            title="Spatial CCI plot",
            x_range=(0, self.dim - 150),
            y_range=(self.dim, 0),
            output_backend=self.output_backend.value,
        )

        fig.image_rgba(
            image=[self.image],
            x=0,
            y=self.xdim,
            dw=self.ydim,
            dh=self.xdim,
            global_alpha=self.tissue_alpha.value,
        )

        # Get query clusters
        selected = self.annot_select.value.strip("raw_")
        command = []
        for i in self.list_cluster.active:
            command.append(
                selected + ' == "' + self.adata[0].obs[selected].cat.categories[i] + '"'
            )
        tmp = self.adata[0].obs.query(" or ".join(command))

        tmp_adata = self.adata[0][tmp.index, :]

        x = tmp_adata.obsm["spatial"][:, 0] * self.scale_factor
        y = tmp_adata.obsm["spatial"][:, 1] * self.scale_factor

        category_items = self.adata[0].obs[selected].cat.categories
        palette = self.adata[0].uns[selected + "_colors"]
        colormap = dict(zip(category_items, palette))
        color = list(tmp[selected].map(colormap))
        cluster = list(tmp[selected])

        s1 = ColumnDataSource(data=dict(x=x, y=y, color=color, cluster=cluster))
        if len(category_items[0]) > 5:
            fig.scatter(
                x="x",
                y="y",
                source=s1,
                size=self.spot_size.value,
                color="color",
                # legend_group="cluster",
                fill_alpha=self.data_alpha.value,
                line_alpha=self.data_alpha.value,
            )
        else:
            fig.scatter(
                x="x",
                y="y",
                source=s1,
                size=self.spot_size.value,
                color="color",
                legend_group="cluster",
                fill_alpha=self.data_alpha.value,
                line_alpha=self.data_alpha.value,
            )

        #### Adding in the arrows for the interaction edges !!! #####
        forward_edges, reverse_edges = self._get_cci_lr_edges()
        self._add_edges(
            fig, self.adata[0], forward_edges, self.arrow_size.value, forward=True
        )
        self._add_edges(
            fig, self.adata[0], reverse_edges, self.arrow_size.value, forward=False
        )

        fig.toolbar.logo = None
        fig.xaxis.visible = False
        fig.yaxis.visible = False
        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.outline_line_alpha = 0
        fig.add_tools(LassoSelectTool())
        fig.add_tools(ZoomOutTool())
        fig.add_tools(HoverTool())
        fig.add_tools(BoxSelectTool())

        hover = fig.select(dict(type=HoverTool))
        hover.tooltips = OrderedDict(
            [
                ("Spot", "$index"),
                ("X location", "@x{1.11}"),
                ("Y location", "@y{1.11}"),
                ("Cluster", "@cluster"),
            ]
        )

        return fig

    def update_data(self, attrname, old, new):
        self.layout.children[1] = self.make_fig()

    def _get_cci_lr_edges(self):
        """Gets edge list of significant interactions for LR pair."""

        adata = self.adata[0]
        lr = self.lr_select.value
        selected = self.annot_select.value

        # Extracting the data #
        l, r = lr.split("_")
        lr_index = np.where(adata.uns["lr_summary"].index.values == lr)[0][0]
        L_bool = adata[:, l].X.toarray()[:, 0] > 0
        R_bool = adata[:, r].X.toarray()[:, 0] > 0
        sig_bool = adata.obsm["lr_sig_scores"][:, lr_index] > 0
        int_df = adata.uns[f"per_lr_cci_{selected}"][lr]

        ###### Getting the edges, from sig-L->R and sig-R<-L ##########
        forward_edges, reverse_edges = get_edges(adata, L_bool, R_bool, sig_bool)

        ###### Subsetting to cell types with significant interactions ########
        spot_bcs = adata.obs_names.values.astype(str)
        spot_labels = adata.obs[selected].values.astype(str)
        label_set = int_df.index.values.astype(str)
        interact_bool = int_df.values > 0

        ###### Subsetting to only significant CCIs ########
        edges_sub = [[], []]  # forward, reverse
        # list re-capitulates edge-counts.
        for i, edges in enumerate([forward_edges, reverse_edges]):
            for j, edge in enumerate(edges):
                k_ = [0, 1] if i == 0 else [1, 0]
                celltype0 = np.where(label_set == spot_labels[spot_bcs == edge[k_[0]]])[
                    0
                ][0]
                celltype1 = np.where(label_set == spot_labels[spot_bcs == edge[k_[1]]])[
                    0
                ][0]
                celltypes = np.array([celltype0, celltype1])
                if interact_bool[celltypes[k_[0]], celltypes[k_[1]]]:
                    edges_sub[i].append(edge)

        return edges_sub

    @staticmethod
    def _add_edges(fig, adata, edges, arrow_size, forward=True, scale_factor=1):
        """Gets edges for input."""
        for i, edge in enumerate(edges):
            cols = ["imagecol", "imagerow"]
            if forward:
                edge0, edge1 = edge
            else:
                edge0, edge1 = edge[::-1]

            # Arrow details #
            x1, y1 = adata.obs.loc[edge0, cols].values.astype(float) * scale_factor
            x2, y2 = adata.obs.loc[edge1, cols].values.astype(float) * scale_factor

            fig.add_layout(
                Arrow(
                    end=VeeHead(size=arrow_size),
                    line_color="black",
                    x_start=x1,
                    y_start=y1,
                    x_end=x2,
                    y_end=y2,
                )
            )

    def update_list(self, attrname, old, name):

        # Initialize the color
        from stlearn.plotting.cluster_plot import cluster_plot

        selected = self.annot_select.value.strip("raw_")
        cluster_plot(self.adata[0], use_label=selected, show_plot=False)

        # self.list_cluster = CheckboxGroup(
        #     labels=list(self.adata[0].obs[self.use_label.value].cat.categories),
        #     active=list(
        #         np.array(range(0, len(self.adata[0].obs[self.use_label.value].unique())))
        #     ),
        # )
        self.list_cluster.labels = list(self.adata[0].obs[selected].cat.categories)
        self.list_cluster.active = list(
            np.array(range(0, len(self.adata[0].obs[selected].unique())))
        )


class Annotate(Spatial):
    def __init__(
        self,
        # plotting param
        adata: AnnData,
    ):
        super().__init__(adata)
        # Open image, and make sure it's RGB*A*
        image = (self.img * 255).astype(np.uint8)

        img_pillow = Image.fromarray(image).convert("RGBA")

        self.xdim, self.ydim = img_pillow.size

        # Create an array representation for the image `img`, and an 8-bit "4
        # layer/RGBA" version of it `view`.
        self.image = np.empty((self.ydim, self.xdim), dtype=np.uint32)
        view = self.image.view(dtype=np.uint8).reshape((self.ydim, self.xdim, 4))
        # Copy the RGBA image into view, flipping it so it comes right-side up
        # with a lower-left origin
        view[:, :, :] = np.flipud(np.asarray(img_pillow))

        # Display the 32-bit RGBA image
        self.dim = max(self.xdim, self.ydim)

        self.data_alpha = Slider(
            title="Spot alpha", value=1.0, start=0, end=1.0, step=0.1
        )

        self.tissue_alpha = Slider(
            title="Tissue alpha", value=1.0, start=0, end=1.0, step=0.1
        )

        self.spot_size = Slider(
            title="Spot size", value=5.0, start=0, end=5.0, step=1.0
        )

        self.checkbox_group = CheckboxGroup(
            labels=["Show spatial trajectories"], active=[]
        )

        inputs = column(
            self.data_alpha,
            self.tissue_alpha,
            self.spot_size,
        )
        self.layout = row(inputs, self.make_fig())

        def modify_fig(doc):
            doc.add_root(row(self.layout, width=800))
            self.data_alpha.on_change("value", self.update_data)
            self.tissue_alpha.on_change("value", self.update_data)
            self.spot_size.on_change("value", self.update_data)

        handler = FunctionHandler(modify_fig)
        self.app = Application(handler)

    def update_data(self, attrname, old, new):
        self.layout.children[1] = self.make_fig()

    def make_fig(self):
        fig = figure(
            x_range=(0, self.dim - 150), y_range=(self.dim, 0), output_backend="webgl"
        )

        colors = np.repeat("black", len(self.imagecol))
        adding_colors = np.array(Category20[20])[
            list(sorted(list(range(0, 20)), key=lambda x: [x % 2, x]))
        ]
        s1 = ColumnDataSource(
            data=dict(x=self.imagecol, y=self.imagerow, colors=colors)
        )

        fig.image_rgba(
            image=[self.image],
            x=0,
            y=self.xdim,
            dw=self.ydim,
            dh=self.xdim,
            global_alpha=self.tissue_alpha.value,
        )

        fig.circle(
            "x",
            "y",
            size=self.spot_size.value,
            color="colors",
            source=s1,
            fill_alpha=self.data_alpha.value,
            line_alpha=self.data_alpha.value,
        )

        fig.toolbar.logo = None
        fig.xaxis.visible = False
        fig.yaxis.visible = False
        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.outline_line_alpha = 0
        fig.add_tools(LassoSelectTool())
        fig.add_tools(BoxSelectTool())

        self.s2 = ColumnDataSource(data=dict(spot=[], label=[]))
        columns = [
            TableColumn(field="spot", title="Spots"),
            TableColumn(field="label", title="Labels"),
        ]
        table = DataTable(
            source=self.s2,
            columns=columns,
            width=400,
            height=200,
            reorderable=False,
            editable=True,
        )

        color_index = ColumnDataSource(data=dict(index=[0]))
        savebutton = Button(label="Add new group", button_type="success", width=200)
        savebutton.js_on_click(
            CustomJS(
                args=dict(
                    source_data=s1,
                    source_data_2=self.s2,
                    adding_colors=adding_colors,
                    color_index=color_index,
                    table=table,
                ),
                code="""

                function addRowToAccumulator(accumulator, spots, labels, index) {
                    accumulator['spot'][index] = spots;
                    accumulator['label'][index] = labels;

                    return accumulator;
                }

                var inds = source_data.selected.indices;
                var data = source_data.data;
                var add_color = adding_colors
                var colors = source_data.data.colors
                var i = 0;
                for (i; i < inds.length; i++) {
                    colors[inds[i]] = add_color[color_index.data.index[0]]
                }

                source_data.change.emit();

                var new_data =  source_data_2.data;

                new_data = addRowToAccumulator(new_data,inds,color_index.data.index[0].toString(),color_index.data.index[0])

                source_data_2.data = new_data

                source_data_2.change.emit();
                color_index.data.index[0] += 1
                color_index.change.emit();
                """,
            )
        )
        submitbutton = Button(label="Submit", button_type="success", width=200)

        def change_click():
            self.adata[0].uns["annotation"] = self.s2.to_df()
            empty_array = np.empty(len(self.adata[0]))
            empty_array[:] = np.NaN
            empty_array = empty_array.astype(object)
            for i in range(0, len(self.adata[0].uns["annotation"])):
                empty_array[
                    [np.array(self.adata[0].uns["annotation"]["spot"][i])]
                ] = str(self.adata[0].uns["annotation"]["label"][i])

            empty_array = pd.Series(empty_array).fillna("other")
            self.adata[0].obs["annotation"] = pd.Categorical(empty_array)

        submitbutton.on_click(change_click)
        submitbutton.js_on_click(
            CustomJS(
                args={},
                code="""
                alert("The annotated labels stored in adata.obs.annotation")
                try {
                        $( "#cluster_plot" ).removeClass( "disabled" )
                    }
                    catch (e) {}
                """,
            )
        )

        layout = column([fig, row(table, column(savebutton, submitbutton))])

        return layout



import numpy as np
import pandas as pd
from anndata import AnnData


def merge(
    adata: AnnData,
    use_lr: str = "cci_lr",
    use_het: str = "cci_het",
    verbose: bool = True,
) -> AnnData:
    """Merge results from cell type heterogeneity and L-R cluster
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    use_lr: str             CCI LR scores
    use_het: str            CCI HET scores

    Returns
    -------
    adata: AnnData          With merged result stored in adata.uns['merged']
    """

    adata.obsm["merged"] = np.multiply(adata.obsm[use_het], adata.obsm[use_lr])

    if verbose:
        print(
            "Results of spatial interaction analysis has been written to adata.uns['merged']"
        )

    return adata



def checkType(arr, n=2):

    # If the first two and the last two elements
    # of the array are in increasing order
    if arr[0] <= arr[1] and arr[n - 2] <= arr[n - 1]:
        return True

    # If the first two and the last two elements
    # of the array are in decreasing order
    elif arr[0] >= arr[1] and arr[n - 2] >= arr[n - 1]:
        return False



# from .base import lr
# from .base_grouping import get_hotspots
# from . import het
# from .het import edge_core, get_between_spot_edge_array
# from .merge import merge
# from .permutation import get_rand_pairs
from .analysis import load_lrs, grid, run, adj_pvals, run_lr_go, run_cci



"""
Helper functions for het.py; primarily counting help.
"""

import numpy as np
from numba.typed import List
from numba import njit


@njit
def edge_core(
    cell_data: np.ndarray,
    cell_type_index: int,
    neighbourhood_bcs: List,
    neighbourhood_indices: List,
    spot_indices: np.array = None,
    neigh_bool: np.array = None,
    cutoff: float = 0.2,
) -> np.array:
    """Gets the edges which connect inputted spots to neighbours of a given cell type.

        Parameters
        ----------
        cell_data: np.ndarray          Spots*CellTypes; value indicates \
                                        proportion of spot due to a specific \
                                        cell type. Rows sum to 1; pure spots \
                                        or spot annotations have a single 1 \
                                        per row.

        cell_type_index: int            Column of cell_data that contains the \
                                        cell type of interest.

        neighbourhood_bcs: List         List of lists, inner list for each \
                                        spot. First element of inner list is \
                                        spot barcode, second element is array \
                                        of neighbourhood spot barcodes.

        neighbourhood_indices: List     Same structure as neighbourhood_bcs, \
                                        but first inner list element is index \
                                        of the spot, and second is array of \
                                        neighbour indices.

        spot_indices: np.array          Array of indices indicating which spots \
                                        we want edges associated with. Can be \
                                        used to subset to significant spots.

        neigh_bool: np.array            Array of booleans of length n-spots, \
                                        True indicates the spot is an allowed \
                                        neighbour. This is useful where we only \
                                        want edges to neighbours which express \
                                        an indicated ligand or receptor.

        cutoff: float                   Cutoff above which cell is considered \
                                        to be present within the spot, is \
                                        applied on cell_data, and thereby allows \
                                        a spot to be counted as having multiple \
                                        cell types.

        Returns
        -------
        edges: List   List of 2-tuples containing spot barcodes, indicating \
                        an edge between a spot in spot_indices and a neighbour \
                        where neigh_bool is True, and either the
    """

    # Subsetting to relevant cell types #
    cell_data = cell_data[:, cell_type_index]

    # Initialising the edge list #
    edge_list = init_edge_list(neighbourhood_bcs)

    # If spots have no neighbours, no counting !
    if len(edge_list) == 0:
        return edge_list
    elif len(spot_indices) == 0:
        return edge_list[1:]

    ### Within-spot mode
    # within-spot, will have only itself as a neighbour in this mode
    within_mode = edge_list[0][0] == edge_list[0][1]
    if within_mode:
        # Numba implimentation #
        for i in spot_indices:
            if neigh_bool[i] and cell_data[i] > cutoff:
                edge_list.append((neighbourhood_bcs[i][0], neighbourhood_bcs[i][1][0]))

    ### Between-spot mode
    else:
        # Subsetting the neighbourhoods to relevant spots #
        neighbourhood_bcs_sub = List()
        neighbourhood_indices_sub = List()
        for spot_i in spot_indices:
            neighbourhood_bcs_sub.append(neighbourhood_bcs[spot_i])
            neighbourhood_indices_sub.append(neighbourhood_indices[spot_i])

        # Number of unique edges will be the count of interactions.
        get_between_spot_edge_array(
            edge_list,
            neighbourhood_bcs_sub,
            neighbourhood_indices_sub,
            neigh_bool,
            cell_data,
            cutoff=cutoff,
        )

    return edge_list[1:]  # Removing the initial edge added for typing #


@njit
def init_edge_list(neighbourhood_bcs):
    """ Initialises the edge-list in a way which ensures consistent typing to \
        prevent errors with Numba.
    """

    # Initialising the edge list #
    edge_list = List()
    # This ensures consistent typing #
    # spots_have_neighbours = False
    # first_spot = -1
    for i in range(len(neighbourhood_bcs)):
        if len(neighbourhood_bcs[i][1]) > 0:
            edge_list.append((neighbourhood_bcs[i][0], neighbourhood_bcs[i][1][0]))
            # spots_have_neighbours = True
            # first_spot = i
            break
    return edge_list


@njit
def get_between_spot_edge_array(
    edge_list: List,
    neighbourhood_bcs: List,
    neighbourhood_indices: List,
    neigh_bool: np.array,
    cell_data: np.array,
    cutoff: float = 0,
):
    """ Populates edge_list with edges linking spots with a valid neighbour \
        of a given cell type. Validity of neighbour determined by neigh_bool, \
        which can indicate whether the neighbour expresses a certain ligand \
        or receptor. See edge_core for parameter information.
    """
    edge_starts = List()
    edge_ends = List()
    for i in range(len(neighbourhood_bcs)):
        bcs, indices = neighbourhood_bcs[i], neighbourhood_indices[i]
        spot_bc, neigh_bcs = bcs
        neigh_indices = indices[1]
        # Subset the neighbours to only those fitting indicated criteria #
        neigh_bcs = neigh_bcs[neigh_bool[neigh_indices]]
        neigh_indices = neigh_indices[neigh_bool[neigh_indices]]

        if len(neigh_indices) == 0:  # No cases where neighbours meet criteria
            continue  # Don't add any interactions for this neighbourhood

        # Note that can keep all by inputting cell_data with all 1's #
        interact_neigh_bool = cell_data[neigh_indices] > cutoff

        # Retrieving the barcodes of the interacting neighbours #
        interact_neigh_bcs = neigh_bcs[interact_neigh_bool]
        for interact_neigh_bc in interact_neigh_bcs:
            edge_starts.append(spot_bc)
            edge_ends.append(interact_neigh_bc)

    # Getting the unique edges #
    if len(edge_starts) > 0:
        add_unique_edges(edge_list, edge_starts, edge_ends)


@njit
def add_unique_edges(edge_list, edge_starts, edge_ends):
    """ Adds the unique edges to the given edge list. \
    Complicated in order to satisfy Numba compilation in no-python mode.
    """
    n_edges = len(edge_starts)

    # Adding the unique edges #
    edge_added = np.zeros((1, len(edge_starts)))[0, :] == 1
    for i in range(n_edges):
        if not edge_added[i]:
            edge_start, edge_end = edge_starts[i], edge_ends[i]
            edge_list.append((edge_start, edge_end))
            for j in range(i, n_edges):
                edge_startj, edge_endj = edge_starts[j], edge_ends[j]
                # Direction doesn't matter #
                if (edge_start == edge_startj and edge_end == edge_endj) or (
                    edge_end == edge_startj and edge_start == edge_endj
                ):
                    edge_added[j] = True


def get_data_for_counting(adata, use_label, mix_mode, all_set):
    """Retrieves the minimal information necessary to perform edge counting."""
    # First determining how the edge counting needs to be performed #
    # Ensuring compatibility with current way of adding label_transfer to object
    if use_label == "label_transfer" or use_label == "predictions":
        obs_key, uns_key = "predictions", "label_transfer"
    else:
        obs_key, uns_key = use_label, use_label

    # Getting the neighbourhoods #
    neighbours, neighbourhood_bcs, neighbourhood_indices = get_neighbourhoods(adata)

    # Getting the cell type information; if not mixtures then populate
    # matrix with one's indicating pure spots.
    if mix_mode:
        cell_props = adata.uns[uns_key]
        cols = cell_props.columns.values.astype(str)
        col_order = [
            np.where([cell_type in col for col in cols])[0][0] for cell_type in all_set
        ]
        cell_data = adata.uns[uns_key].iloc[:, col_order].values.astype(np.float64)
    else:
        cell_labels = adata.obs.loc[:, obs_key].values
        cell_data = np.zeros((len(cell_labels), len(all_set)), dtype=np.float64)
        for i, cell_type in enumerate(all_set):
            cell_data[:, i] = (
                (cell_labels == cell_type).astype(np.int_).astype(np.float64)
            )

    spot_bcs = adata.obs_names.values.astype(str)
    return spot_bcs, cell_data, neighbourhood_bcs, neighbourhood_indices


def get_neighbourhoods(adata):
    """Gets the neighbourhood information."""

    # Old stlearn version where didn't store neighbourhood barcodes, not good
    #   for anndata subsetting!!
    if not "spot_neigh_bcs" in adata.obsm:
        # Determining the neighbour spots used for significance testing #
        neighbours = List()
        for i in range(adata.obsm["spot_neighbours"].shape[0]):
            neighs = np.array(adata.obsm["spot_neighbours"].values[i, :][0].split(","))
            neighs = neighs[neighs != ""].astype(int)
            # DONE: store neigh_bcs to get below to work reliably
            #       after subsetting anndata #
            # neighs = neighs[neighs<adata.shape[0]] # Removing subsetted spots..
            neighbours.append(neighs)

        # Getting the neighbourhood information #
        neighbourhood_bcs = List()
        neighbourhood_indices = List()
        spot_bcs = adata.obs_names.values.astype(str)
        for spot_i in range(len(spot_bcs)):
            neighbourhood_indices.append((spot_i, neighbours[spot_i]))
            neighbourhood_bcs.append((spot_bcs[spot_i], spot_bcs[neighbours[spot_i]]))
    else:  # Newer version
        # Determining the neighbour spots used for significance testing #
        neighbours = List()
        neighbourhood_bcs = List()
        neighbourhood_indices = List()
        spot_bcs = adata.obs_names.values.astype(str)
        str_dtype = f"<U{max([len(bc) for bc in spot_bcs])}"  # ensures correct typing
        for i in range(adata.shape[0]):
            neigh_bcs = np.array(
                adata.obsm["spot_neigh_bcs"].values[i, :][0].split(",")
            )
            neigh_bcs = neigh_bcs[neigh_bcs != ""]
            neigh_bcs = np.array(
                [neigh_bc for neigh_bc in neigh_bcs if neigh_bc in spot_bcs],
                dtype=str_dtype,
            )
            neigh_indices = np.array(
                [np.where(spot_bcs == neigh_bc)[0][0] for neigh_bc in neigh_bcs],
                dtype=np.int64,
            )
            neighbours.append(neigh_indices)
            neighbourhood_indices.append((i, neigh_indices))
            neighbourhood_bcs.append((spot_bcs[i], neigh_bcs))

    return neighbours, neighbourhood_bcs, neighbourhood_indices



from typing import Optional
import histomicstk as htk
import numpy as np
import scipy as sp
import skimage.color
import skimage.io
import skimage.measure
from anndata import AnnData
from scipy import ndimage as ndi
from skimage.feature import peak_local_max
from skimage.segmentation import watershed
from tqdm import tqdm


def morph_watershed(
    adata: AnnData,
    library_id: str = None,
    verbose: bool = False,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Watershed method to segment nuclei and calculate morphological statistics

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **n_nuclei** : `adata.obs` field
        saved number of nuclei of each spot image tiles
    **nuclei_total_area** : `adata.obs` field
        saved of total area of nuclei of each spot image tiles
    **nuclei_mean_area** : `adata.obs` field
        saved mean area of nuclei of each spot image tiles
    **nuclei_std_area** : `adata.obs` field
        saved stand deviation of nuclei area of each spot image tiles
    **eccentricity** : `adata.obs` field
        saved eccentricity of each spot image tiles
    **mean_pix_r** : `adata.obs` field
        saved mean pixel value of red channel of of each spot image tiles
    **std_pix_r** : `adata.obs` field
        saved stand deviation of red channel of each spot image tiles
    **mean_pix_g** : `adata.obs` field
        saved mean pixel value of green channel of each spot image tiles
    **std_pix_g** : `adata.obs` field
        saved stand deviation of green channel of each spot image tiles
    **mean_pix_b** : `adata.obs` field
        saved mean pixel value of blue channel of each spot image tiles
    **std_pix_b** : `adata.obs` field
        saved stand deviation of blue channel of each spot image tiles
    **nuclei_total_area_per_tile** : `adata.obs` field
        saved total nuclei area per tile of each spot image tiles
    """

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    n_nuclei_list = []
    nuclei_total_area_list = []
    nuclei_mean_area_list = []
    nuclei_std_area_list = []
    eccentricity_list = []
    mean_pix_list_r = []
    std_pix_list_r = []
    mean_pix_list_g = []
    std_pix_list_g = []
    mean_pix_list_b = []
    std_pix_list_b = []
    with tqdm(
        total=len(adata),
        desc="calculate morphological stats",
        bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:
        for tile in adata.obs["tile_path"]:
            (
                n_nuclei,
                nuclei_total_area,
                nuclei_mean_area,
                nuclei_std_area,
                eccentricity,
                solidity,
                mean_pix_r,
                std_pix_r,
                mean_pix_g,
                std_pix_g,
                mean_pix_b,
                std_pix_b,
            ) = _calculate_morph_stats(tile)
            n_nuclei_list.append(n_nuclei)
            nuclei_total_area_list.append(nuclei_total_area)
            nuclei_mean_area_list.append(nuclei_mean_area)
            nuclei_std_area_list.append(nuclei_std_area)
            eccentricity_list.append(eccentricity)
            mean_pix_list_r.append(mean_pix_r)
            std_pix_list_r.append(std_pix_r)
            mean_pix_list_g.append(mean_pix_g)
            std_pix_list_g.append(std_pix_g)
            mean_pix_list_b.append(mean_pix_b)
            std_pix_list_b.append(std_pix_b)
            pbar.update(1)

    adata.obs["n_nuclei"] = n_nuclei_list
    adata.obs["nuclei_total_area"] = nuclei_total_area_list
    adata.obs["nuclei_mean_area"] = nuclei_mean_area_list
    adata.obs["nuclei_std_area"] = nuclei_std_area_list
    adata.obs["eccentricity"] = eccentricity_list
    adata.obs["mean_pix_r"] = mean_pix_list_r
    adata.obs["std_pix_r"] = std_pix_list_r
    adata.obs["mean_pix_g"] = mean_pix_list_g
    adata.obs["std_pix_g"] = std_pix_list_g
    adata.obs["mean_pix_b"] = mean_pix_list_b
    adata.obs["std_pix_b"] = std_pix_list_b
    adata.obs["nuclei_total_area_per_tile"] = adata.obs["nuclei_total_area"] / 299 / 299
    return adata if copy else None


def _calculate_morph_stats(tile_path):
    imInput = skimage.io.imread(tile_path)
    stain_color_map = htk.preprocessing.color_deconvolution.stain_color_map
    stains = [
        "hematoxylin",  # nuclei stain
        "eosin",  # cytoplasm stain
        "null",
    ]  # set to null if input contains only two stains
    w_est = htk.preprocessing.color_deconvolution.rgb_separate_stains_macenko_pca(
        imInput, 255
    )

    # Perform color deconvolution
    deconv_result = htk.preprocessing.color_deconvolution.color_deconvolution(
        imInput, w_est, 255
    )

    channel = htk.preprocessing.color_deconvolution.find_stain_index(
        stain_color_map[stains[0]], w_est
    )
    im_nuclei_stain = deconv_result.Stains[:, :, channel]

    thresh = skimage.filters.threshold_otsu(im_nuclei_stain)
    # im_fgnd_mask = im_nuclei_stain < thresh
    im_fgnd_mask = sp.ndimage.morphology.binary_fill_holes(
        im_nuclei_stain < 0.8 * thresh
    )

    distance = ndi.distance_transform_edt(im_fgnd_mask)
    coords = peak_local_max(distance, footprint=np.ones((3, 3)), labels=im_fgnd_mask)
    mask = np.zeros(distance.shape, dtype=bool)
    mask[tuple(coords.T)] = True
    markers, _ = ndi.label(mask)

    labels = watershed(im_nuclei_stain, markers, mask=im_fgnd_mask)
    min_nucleus_area = 60
    im_nuclei_seg_mask = htk.segmentation.label.area_open(
        labels, min_nucleus_area
    ).astype(np.int)

    # compute nuclei properties
    objProps = skimage.measure.regionprops(im_nuclei_seg_mask)

    #     # Display results
    #     plt.figure(figsize=(20, 10))
    #     plt.imshow(skimage.color.label2rgb(im_nuclei_seg_mask, im_nuclei_stain, bg_label=0),
    #            origin='upper')
    #     plt.title('Nuclei segmentation mask overlay')
    #     plt.savefig("./Nuclei_segmentation_tiles_bc_wh/{}.png".format(tile_path.split("/")[-1].split(".")[0]), dpi=300)

    n_nuclei = len(objProps)

    nuclei_total_area = sum(map(lambda x: x.area, objProps))
    nuclei_mean_area = np.mean(list(map(lambda x: x.area, objProps)))
    nuclei_std_area = np.std(list(map(lambda x: x.area, objProps)))

    mean_pix = imInput.reshape(3, -1).mean(1)
    std_pix = imInput.reshape(3, -1).std(1)

    eccentricity = np.mean(list(map(lambda x: x.eccentricity, objProps)))

    solidity = np.mean(list(map(lambda x: x.solidity, objProps)))

    return (
        n_nuclei,
        nuclei_total_area,
        nuclei_mean_area,
        nuclei_std_area,
        eccentricity,
        solidity,
        mean_pix[0],
        std_pix[0],
        mean_pix[1],
        std_pix[1],
        mean_pix[2],
        std_pix[2],
    )



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
import math
import random
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
import io
from copy import deepcopy
from stlearn.utils import _read_graph


def tree_plot(
    adata: AnnData,
    library_id: str = None,
    figsize: Union[float, int] = (10, 4),
    data_alpha: float = 1.0,
    use_label: str = "louvain",
    spot_size: Union[float, int] = 50,
    fontsize: int = 6,
    piesize: float = 0.15,
    zoom: float = 0.1,
    name: str = None,
    output: str = None,
    dpi: int = 180,
    show_all: bool = False,
    show_plot: bool = True,
    ncols: int = 4,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Hierarchical tree plot represent for the global spatial trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of cluster method.
    figsize
        Change figure size.
    data_alpha
        Opacity of the spot.
    fontsize
        Choose font size.
    piesize
        Choose the size of cropped image.
    zoom
        Choose zoom factor.
    show_all
        Show all cropped image or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    G = _read_graph(adata, "PTS_graph")

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    G.remove_node(9999)

    start_nodes = []
    disconnected_nodes = []
    for node in G.in_degree():
        if node[1] == 0:
            start_nodes.append(node[0])

    for node in G.out_degree():
        if node[1] == 0:
            disconnected_nodes.append(node[0])

    start_nodes = list(set(start_nodes) - set(disconnected_nodes))
    start_nodes.sort()

    nrows = math.ceil(len(start_nodes) / ncols)

    superfig, axs = plt.subplots(nrows, ncols, figsize=figsize)
    axs = axs.ravel()

    for idx in range(0, nrows * ncols):
        try:
            generate_tree_viz(
                adata, use_label, G, axs[idx], starter_node=start_nodes[idx]
            )
        except:
            axs[idx] = axs[idx].axis("off")

    if name is None:
        name = use_label

    if output is not None:
        superfig.savefig(
            output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0
        )

    if show_plot == True:
        plt.show()


def hierarchy_pos(G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5):

    """
    From Joel's answer at https://stackoverflow.com/a/29597209/2966723.
    Licensed under Creative Commons Attribution-Share Alike

    If the graph is a tree this will return the positions to plot this in a
    hierarchical layout.

    G: the graph (must be a tree)

    root: the root node of current branch
    - if the tree is directed and this is not given,
      the root will be found and used
    - if the tree is directed and this is given, then
      the positions will be just for the descendants of this node.
    - if the tree is undirected and not given,
      then a random choice will be used.

    width: horizontal space allocated for this branch - avoids overlap with other branches

    vert_gap: gap between levels of hierarchy

    vert_loc: vertical location of root

    xcenter: horizontal location of root
    """
    if not nx.is_tree(G):
        raise TypeError("cannot use hierarchy_pos on a graph that is not a tree")

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(
                iter(nx.topological_sort(G))
            )  # allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(
        G, root, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5, pos=None, parent=None
    ):
        """
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        """

        if pos is None:
            pos = {root: (xcenter, vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children) != 0:
            dx = width / len(children)
            nextx = xcenter - width / 2 - dx / 2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(
                    G,
                    child,
                    width=dx,
                    vert_gap=vert_gap,
                    vert_loc=vert_loc - vert_gap,
                    xcenter=nextx,
                    pos=pos,
                    parent=root,
                )
        return pos

    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


def generate_tree_viz(adata, use_label, G, axis, starter_node):
    tmp_edges = []
    for edge in G.edges():
        if starter_node == edge[0]:
            tmp_edges.append(edge)
    tmp_D = nx.DiGraph()
    tmp_D.add_edges_from(tmp_edges)

    pos = hierarchy_pos(tmp_D)
    a = axis

    a.axis("off")
    colors = []
    for n in tmp_D:
        subset = adata.obs[adata.obs["sub_cluster_labels"] == str(n)]
        colors.append(adata.uns[use_label + "_colors"][int(subset[use_label][0])])

    nx.draw_networkx_edges(
        tmp_D,
        pos,
        ax=a,
        arrowstyle="-",
        edge_color="#ADABAF",
        connectionstyle="angle3,angleA=0,angleB=90",
    )
    nx.draw_networkx_nodes(tmp_D, pos, node_color=colors, ax=a)
    nx.draw_networkx_labels(tmp_D, pos, font_color="black", ax=a)



""" Helper functions for views.py.
"""

import numpy


def getVal(form, element):
    return getattr(form, element).data


def getData(form):
    """Retrieves the data from the form and places into dictionary."""
    params = {}
    form_elements = form.elements
    form_fields = form.element_fields
    for i, element in enumerate(form_elements):
        if form_fields[i] != "Title":
            data = getVal(form, element)
            params[element] = data
    return params


def getLR(lr_input, gene_names):
    """Returns list of lr_inputs and error message, if any."""
    if lr_input == "":
        return None, "ERROR: LR pairs required input."

    try:
        lrs = [lr.strip(" ") for lr in lr_input.split(",")]
        absent_genes = []
        for lr in lrs:
            genes = lr.split("_")
            absent_genes.extend([gene for gene in genes if gene not in gene_names])

        if len(absent_genes) != 0:
            return None, f"ERROR: inputted genes not found {absent_genes}."

        return lrs, ""

    except:
        return None, "ERROR: LR pairs misformatted."



#!/usr/bin/env python

"""Tests for `stlearn` package."""

import os

import unittest

import numpy as np
from numba.typed import List

import stlearn as st
import scanpy as sc
from tests.utils import read_test_data

import stlearn.tools.microenv.cci.het_helpers as het_hs
import stlearn.tools.microenv.cci.het as het

global adata
adata = read_test_data()


class TestCCI(unittest.TestCase):
    """Tests for `stlearn` CCI capability."""

    def setUp(self) -> None:
        """Setup some basic test-cases as sanity checks."""

        ##### Unit neighbourhood, containing just 1 spot and 6 neighbours ######
        """
        * A is the middle spot, B/C/D/E/F/G are the neighbouring spots clock-
                                                  wise starting at the top-left.
        """
        spot_bcs = np.array(["A", "B", "C", "D", "E", "F", "G"])
        neigh_dict = {
            "A": ["B", "C", "D", "E", "F", "G"],
            "B": ["G", "A", "C"],
            "C": ["B", "A", "D"],
            "D": ["E", "A", "C"],
            "E": ["F", "A", "D"],
            "F": ["G", "A", "E"],
            "G": ["F", "A", "B"],
        }
        neighbourhood_bcs = List()
        neighbourhood_indices = List()
        for i, bc in enumerate(spot_bcs):
            neigh_bcs = np.array(neigh_dict[bc])
            neigh_indices = np.array(
                [np.where(spot_bcs == n_bc)[0][0] for n_bc in neigh_bcs]
            )
            neighbourhood_bcs.append((bc, neigh_bcs))
            neighbourhood_indices.append((i, neigh_indices))

        self.neighbourhood_bcs = neighbourhood_bcs
        self.neighbourhood_indices = neighbourhood_indices
        self.neigh_dict = neigh_dict

    ##### Basic tests #######
    def test_load_lrs(self):
        """Testing loading lr database."""
        sizes = [2293, 4071]  # lit lr db size, putative lr db size.
        lrs = st.tl.cci.load_lrs(["connectomeDB2020_lit"])
        self.assertEqual(len(lrs), sizes[0])

        lrs = st.tl.cci.load_lrs(["connectomeDB2020_put"])
        self.assertEqual(len(lrs), sizes[1])

        lrs = st.tl.cci.load_lrs(["connectomeDB2020_lit", "connectomeDB2020_put"])
        self.assertEqual(len(lrs), sizes[1])

        lrs = st.tl.cci.load_lrs()
        self.assertEqual(len(lrs), sizes[0])

        ### Testing loading mouse as species ####
        lrs = st.tl.cci.load_lrs(species="mouse")
        genes1 = [lr_.split("_")[0] for lr_ in lrs]
        genes2 = [lr_.split("_")[1] for lr_ in lrs]
        self.assertTrue(np.all([gene[0].isupper() for gene in genes1]))
        self.assertTrue(np.all([gene[1:] == gene[1:].lower() for gene in genes1]))
        self.assertTrue(np.all([gene[0].isupper() for gene in genes2]))
        self.assertTrue(np.all([gene[1:] == gene[1:].lower() for gene in genes2]))

    ####### Important, granular tests related to LR scoring #########

    ###### Important, granular tests related to CCI counting #######
    def test_edge_retrieval_basic(self):
        """ Basic test of functionality to retrieve edges via \
                                                    get_between_spot_edge_array.
        """
        neighbourhood_bcs = self.neighbourhood_bcs
        neighbourhood_indices = self.neighbourhood_indices

        # Initialising the edge list #
        edge_list = het_hs.init_edge_list(neighbourhood_bcs)

        ############# Basic case, should populate with all edges ###############
        neigh_bool = np.array([True] * len(neighbourhood_bcs))
        cell_data = np.array([1] * len(neighbourhood_bcs), dtype=np.float)
        het_hs.get_between_spot_edge_array(
            edge_list, neighbourhood_bcs, neighbourhood_indices, neigh_bool, cell_data
        )
        edge_list = edge_list[1:]  # Remove initialisation edge

        # Getting all the edges in the neighbourhood #
        all_edges = []
        for bc in self.neigh_dict:
            neigh_bcs = self.neigh_dict[bc]
            for neigh_bc in neigh_bcs:
                if (bc, neigh_bc) not in all_edges and (neigh_bc, bc) not in all_edges:
                    all_edges.append((bc, neigh_bc))
        n_edges = len(all_edges)

        self.assertEqual(len(edge_list), n_edges)
        self.assertTrue(
            np.all([edge in all_edges or edge[::-1] in all_edges for edge in edge_list])
        )

        ########### Some neighbours not valid but no effect on edge list #######
        # No effect since though not a valid neighbour, still a valid spot #
        edge_list = het_hs.init_edge_list(neighbourhood_bcs)
        invalid_neighs = ["B", "E"]
        neigh_bool = np.array([bc not in invalid_neighs for bc in self.neigh_dict])
        het_hs.get_between_spot_edge_array(
            edge_list, neighbourhood_bcs, neighbourhood_indices, neigh_bool, cell_data
        )
        edge_list = edge_list[1:]  # Remove initialisation edge

        self.assertEqual(len(edge_list), n_edges)
        self.assertTrue(
            np.all([edge in all_edges or edge[::-1] in all_edges for edge in edge_list])
        )

        ########### Some neighbours not valid, effects the edge list ###########
        # Two neighbouring spots no longer valid neighbours #
        edge_list = het_hs.init_edge_list(neighbourhood_bcs)
        invalid_neighs = ["B", "C"]
        neigh_bool = np.array([bc not in invalid_neighs for bc in self.neigh_dict])
        het_hs.get_between_spot_edge_array(
            edge_list, neighbourhood_bcs, neighbourhood_indices, neigh_bool, cell_data
        )
        edge_list = edge_list[1:]  # Remove initialisation edge

        sub_edges1 = [
            edge for edge in all_edges if edge not in [("B", "C"), ("C", "B")]
        ]
        n_subedges = len(sub_edges1)

        self.assertEqual(len(edge_list), n_subedges)
        self.assertTrue(
            np.all(
                [edge in sub_edges1 or edge[::-1] in sub_edges1 for edge in edge_list]
            )
        )

        ######### Middle spot not neighbour, cell type, or spot of interest ####
        # Removing the centre-spot as being the cell type of interest #
        neigh_bool = np.array([True] * len(neighbourhood_bcs))
        neigh_bool[0] = False
        cell_data = np.array([1] * len(neighbourhood_bcs), dtype=np.float)
        cell_data[0] = 0
        neigh_bcs = neighbourhood_bcs[1:]
        neigh_indices = neighbourhood_indices[1:]

        edge_list = het_hs.init_edge_list(neighbourhood_bcs)
        het_hs.get_between_spot_edge_array(
            edge_list, neigh_bcs, neigh_indices, neigh_bool, cell_data
        )
        edge_list = edge_list[1:]  # Remove initialisation edge

        sub_edges2 = [edge for edge in all_edges if "A" not in edge]
        n_subedges2 = len(sub_edges2)

        self.assertEqual(len(edge_list), n_subedges2)
        self.assertTrue(
            np.all(
                [edge in sub_edges2 or edge[::-1] in sub_edges2 for edge in edge_list]
            )
        )

        ### Corner spot valid neighbour, not cell type, not spot of interest ###
        neigh_bool = np.array([True] * len(neighbourhood_bcs))
        cell_data = np.array([1] * len(neighbourhood_bcs), dtype=np.float)
        cell_data[1] = 0
        neigh_bcs, neigh_indices = List(), List()
        for i in range(len(cell_data)):
            if i != 1:
                neigh_bcs.append(neighbourhood_bcs[i])
                neigh_indices.append(neighbourhood_indices[i])

        edge_list = het_hs.init_edge_list(neighbourhood_bcs)
        het_hs.get_between_spot_edge_array(
            edge_list, neigh_bcs, neigh_indices, neigh_bool, cell_data
        )
        edge_list = edge_list[1:]  # Remove initialisation edge

        sub_edges3 = [edge for edge in all_edges if "B" not in edge]
        n_subedges3 = len(sub_edges3)

        self.assertEqual(len(edge_list), n_subedges3)
        self.assertTrue(
            np.all(
                [edge in sub_edges3 or edge[::-1] in sub_edges3 for edge in edge_list]
            )
        )

    def test_get_interactions(self):
        """Basic test of capability to get interactions betwen cell types when \
            considering spots of a particular cell type expressing a ligand, \
            and spots of another cell type expressing the receptor.
        """

        ####### Case 1 ######
        """ Middle spot only spot of interest.
            Cell type 1, 2, or 3.
            Middle spot expresses ligand.
            3 neighbours express receptor:
                * One is cell type 1, two are cell type 2.
        """
        cell_annots = [1, 2, 3, 2, 1, 3, 2]
        cell_data = np.zeros((len(cell_annots), 3), dtype=np.float)
        for i, annot in enumerate(cell_annots):
            cell_data[i, annot - 1] = 1
        all_set = np.array([str(i) for i in range(1, 4)])
        sig_bool = np.array([True] + ([False] * (len(cell_annots) - 1)))
        l_bool = sig_bool
        r_bool = np.array([False] * len(cell_annots))
        r_bool[[3, 4, 6]] = True

        # NOTE that format of output is an edge list for each celltype-celltype
        # interaction, where edge list represents interactions between:
        #   1-1, 1-2, 1-3, 2-1, 2-2, 2-3, 3-1, 3-2, 3-3
        n_ccis_comp = 3 * 3
        # one interaction between cell type one and itself,
        # two interactions between cell type one and 2, rest no interactions.
        expected_edges = [[("A", "E")], [("A", "D"), ("A", "G")]]
        [expected_edges.append([]) for i in range(n_ccis_comp - 2)]

        interaction_edges = het.get_interactions(
            cell_data,
            self.neighbourhood_bcs,
            self.neighbourhood_indices,
            all_set,
            sig_bool,
            l_bool,
            r_bool,
            0,
        )

        self.assertEqual(len(interaction_edges), len(expected_edges))
        for i in range(len(expected_edges)):
            expect_edgesi = expected_edges[i]
            observed_edgesi = interaction_edges[i]
            match_bool = [
                edge in observed_edgesi or edge[::-1] in observed_edgesi
                for edge in expect_edgesi
            ]
            self.assertEqual(len(observed_edgesi), len(expect_edgesi))
            self.assertTrue(np.all(match_bool))

    # TODO next things to test:
    #   1. Getting the interaction matrix.
    #   2. Getting the LR scores.



from typing import Optional, Union
from anndata import AnnData
import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib as mpl
import numpy as np
import stlearn.plotting.utils as utils


def deconvolution_plot(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    cluster: [int, str] = None,
    celltype: str = None,
    celltype_threshold: float = 0,
    data_alpha: float = 1.0,
    threshold: float = 0.0,
    cmap: str = "tab20",
    colors: list = None,  # The colors to use for each label...
    tissue_alpha: float = 1.0,
    title: str = None,
    spot_size: Union[float, int] = 10,
    show_axis: bool = False,
    show_legend: bool = True,
    show_donut: bool = True,
    cropped: bool = True,
    margin: int = 100,
    name: str = None,
    dpi: int = 150,
    output: str = None,
    copy: bool = False,
    figsize: tuple = (6.4, 4.8),
    show=True,
) -> Optional[AnnData]:

    """\
    Clustering plot for sptial transcriptomics data. Also it has a function to display trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of cluster method.
    list_cluster
        Choose set of clusters that will display in the plot.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    show_donut
        Whether to show the donut plot or not.
    show_trajectory
        Show the spatial trajectory or not. It requires stlearn.spatial.trajectory.pseudotimespace.
    show_subcluster
        Show subcluster or not. It requires stlearn.spatial.trajectory.global_level.
    name
        Name of the output figure file.
    dpi
        DPI of the output figure.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    # plt.rcParams['figure.dpi'] = dpi

    imagecol = adata.obs["imagecol"]
    imagerow = adata.obs["imagerow"]

    fig, ax = plt.subplots(figsize=figsize)

    label = adata.obsm["deconvolution"].T

    tmp = label.sum(axis=1)

    label_filter = label.loc[tmp[tmp >= np.quantile(tmp, threshold)].index]

    if cluster is not None:
        base = adata.obs[adata.obs[use_label] == str(cluster)][["imagecol", "imagerow"]]
    else:
        base = adata.obs[["imagecol", "imagerow"]]

    if celltype is not None:
        base = base.loc[
            adata.obs_names[adata.obsm["deconvolution"][celltype] > celltype_threshold]
        ]

    label_filter_ = label_filter[base.index]

    if type(colors) == type(None):
        color_vals = list(range(0, len(label_filter_), 1))
        my_norm = mpl.colors.Normalize(0, len(label_filter_))
        my_cmap = mpl.cm.get_cmap(cmap, len(color_vals))
        colors = my_cmap.colors

    for i, xy in enumerate(base.values):
        _ = ax.pie(
            label_filter_.T.iloc[i].values,
            colors=colors,
            center=(xy[0], xy[1]),
            radius=spot_size,
            frame=True,
        )
    ax.autoscale()

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][
        adata.uns["spatial"][library_id]["use_quality"]
    ]

    if show_donut:
        ax_pie = fig.add_axes([0.5, -0.4, 0.03, 0.5])

        def my_autopct(pct):
            return ("%1.0f%%" % pct) if pct >= 4 else ""

        ax_pie.pie(
            label_filter_.sum(axis=1),
            colors=my_cmap.colors,
            radius=10,
            # frame=True,
            autopct=my_autopct,
            pctdistance=1.1,
            startangle=90,
            wedgeprops=dict(width=(3), edgecolor="w", antialiased=True),
            textprops={"fontsize": 5},
        )

    if show_legend == True:
        ax_cb = fig.add_axes([0.9, 0.25, 0.03, 0.5], axisbelow=False)
        cb = mpl.colorbar.ColorbarBase(
            ax_cb, cmap=my_cmap, norm=my_norm, ticks=color_vals
        )

        cb.ax.tick_params(size=0)
        loc = np.array(color_vals) + 0.5
        cb.set_ticks(loc)
        cb.set_ticklabels(label_filter_.index)
        cb.outline.set_visible(False)

    # Overlay the tissue image
    ax.imshow(
        image,
        alpha=1,
        zorder=-1,
    )

    ax.axis("off")

    if cropped:
        ax.set_xlim(imagecol.min() - margin, imagecol.max() + margin)

        ax.set_ylim(imagerow.min() - margin, imagerow.max() + margin)

        ax.set_ylim(ax.get_ylim()[::-1])

        # plt.gca().invert_yaxis()

    if name is None:
        name = use_label

    if output is not None:
        fig.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)

    if show:
        plt.show()



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings

from stlearn.utils import _read_graph


def pseudotime_plot(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    pseudotime_key: str = "pseudotime_key",
    list_clusters: Union[str, list] = None,
    cell_alpha: float = 1.0,
    image_alpha: float = 1.0,
    edge_alpha: float = 0.8,
    node_alpha: float = 1.0,
    spot_size: Union[float, int] = 6.5,
    node_size: float = 5,
    show_color_bar: bool = True,
    show_axis: bool = False,
    show_graph: bool = True,
    show_trajectories: bool = False,
    reverse: bool = False,
    show_node: bool = True,
    show_plot: bool = True,
    cropped: bool = True,
    margin: int = 100,
    dpi: int = 150,
    output: str = None,
    name: str = None,
    copy: bool = False,
    ax=None,
) -> Optional[AnnData]:

    """\
    Global trajectory inference plot (Only DPT).

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of cluster method.
    list_clusters
        Choose set of clusters that will display in the plot.
    cell_alpha
        Opacity of the spot.
    image_alpha
        Opacity of the tissue.
    edge_alpha
        Opacity of edge in PAGA graph in the tissue.
    node_alpha
        Opacity of node in PAGA graph in the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    node_size
        Size of node in PAGA graph in the tissue.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_graph
        Show PAGA graph or not.
    show_legend
        Show legend or not.
    show_plot
        Show plot or not
    dpi
        DPI of the output figure.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    # plt.rcParams['figure.dpi'] = dpi

    imagecol = adata.obs["imagecol"]
    imagerow = adata.obs["imagerow"]

    if list_clusters == None:
        list_clusters = np.array(range(0, len(adata.obs[use_label].unique()))).astype(
            int
        )
    # Get query clusters
    command = []
    # for i in list_clusters:
    #    command.append(use_label + ' == "' + str(i) + '"')
    # tmp = adata.obs.query(" or ".join(command))
    tmp = adata.obs
    G = _read_graph(adata, "global_graph")

    labels = nx.get_edge_attributes(G, "weight")

    result = []
    query_node = get_node(list_clusters, adata.uns["split_node"])
    for edge in G.edges(query_node):
        if (edge[0] in query_node) and (edge[1] in query_node):
            result.append(edge)

    result2 = []
    for edge in result:
        try:
            result2.append(labels[edge] + 0.5)
        except:
            result2.append(labels[edge[::-1]] + 0.5)

    fig, a = plt.subplots()
    if ax != None:
        a = ax
    centroid_dict = adata.uns["centroid_dict"]
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}
    dpt = adata.obs[pseudotime_key]

    colors = adata.obs[use_label].astype(int)
    vmin = min(dpt)
    vmax = max(dpt)
    # Plot scatter plot based on pixel of spots
    from sklearn.preprocessing import MinMaxScaler

    scaler = MinMaxScaler()
    scale = scaler.fit_transform(tmp[pseudotime_key].values.reshape(-1, 1)).reshape(
        -1, 1
    )

    plot = a.scatter(
        tmp["imagecol"],
        tmp["imagerow"],
        edgecolor="none",
        alpha=cell_alpha,
        s=spot_size,
        marker="o",
        vmin=vmin,
        vmax=vmax,
        cmap=plt.get_cmap("viridis"),
        c=scale.reshape(1, -1)[0],
    )

    n_clus = len(colors.unique())

    used_colors = adata.uns[use_label + "_colors"]
    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", used_colors)

    cmap = plt.get_cmap(cmaps)
    bounds = np.linspace(0, n_clus, n_clus + 1)
    norm = matplotlib.colors.BoundaryNorm(bounds, cmap.N)

    norm = matplotlib.colors.Normalize(vmin=min(colors), vmax=max(colors))
    m = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)

    if show_graph:
        nx.draw_networkx(
            G,
            pos=centroid_dict,
            node_size=1,
            font_size=0,
            linewidths=2,
            edgelist=result,
            width=result2,
            alpha=edge_alpha,
            edge_color="#333333",
        )

        for x, y in centroid_dict.items():

            if x in get_node(list_clusters, adata.uns["split_node"]):
                a.text(
                    y[0],
                    y[1],
                    get_cluster(str(x), adata.uns["split_node"]),
                    color="white",
                    fontsize=node_size,
                    zorder=100,
                    bbox=dict(
                        facecolor=cmap(
                            int(get_cluster(str(x), adata.uns["split_node"]))
                            / (len(used_colors) - 1)
                        ),
                        boxstyle="circle",
                        alpha=node_alpha,
                    ),
                )

    if show_trajectories:

        used_colors = adata.uns[use_label + "_colors"]
        cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", used_colors)

        cmap = plt.get_cmap(cmaps)

        if "PTS_graph" not in adata.uns:
            raise ValueError("Please run stlearn.spatial.trajectory.pseudotimespace!")

        tmp = _read_graph(adata, "PTS_graph")

        G = tmp.copy()

        remove = [edge for edge in G.edges if 9999 in edge]
        G.remove_edges_from(remove)
        G.remove_node(9999)
        centroid_dict = adata.uns["centroid_dict"]
        centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}
        if reverse:
            nx.draw_networkx_edges(
                G,
                pos=centroid_dict,
                node_size=10,
                alpha=1.0,
                width=2.5,
                edge_color="#f4efd3",
                arrowsize=17,
                arrowstyle="<|-",
                connectionstyle="arc3,rad=0.2",
            )
        else:
            nx.draw_networkx_edges(
                G,
                pos=centroid_dict,
                node_size=10,
                alpha=1.0,
                width=2.5,
                edge_color="#f4efd3",
                arrowsize=17,
                arrowstyle="-|>",
                connectionstyle="arc3,rad=0.2",
            )

        if show_node:
            for x, y in centroid_dict.items():

                if x in get_node(list_clusters, adata.uns["split_node"]):
                    a.text(
                        y[0],
                        y[1],
                        get_cluster(str(x), adata.uns["split_node"]),
                        color="black",
                        fontsize=8,
                        zorder=100,
                        bbox=dict(
                            facecolor=cmap(
                                int(get_cluster(str(x), adata.uns["split_node"]))
                                / (len(used_colors) - 1)
                            ),
                            boxstyle="circle",
                            alpha=1,
                        ),
                    )

    if show_color_bar:
        cb = plt.colorbar(plot, cmap="viridis")
        cb.outline.set_visible(False)

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][
        adata.uns["spatial"][library_id]["use_quality"]
    ]

    a.imshow(
        image,
        alpha=image_alpha,
        zorder=-1,
    )

    if not show_axis:
        a.axis("off")

    if cropped:
        a.set_xlim(imagecol.min() - margin, imagecol.max() + margin)

        a.set_ylim(imagerow.min() - margin, imagerow.max() + margin)

        a.set_ylim(a.get_ylim()[::-1])
        # plt.gca().invert_yaxis()

    if output is not None:
        fig.savefig(output + "/" + name, dpi=dpi, bbox_inches="tight", pad_inches=0)
    if show_plot == True:
        plt.show()


# get name of cluster by subcluster
def get_cluster(search, dictionary):
    for cl, sub in dictionary.items():
        if search in sub:
            return cl


def get_node(node_list, split_node):
    result = np.array([])
    for node in node_list:
        result = np.append(result, np.array(split_node[int(node)]).astype(int))
    return result.astype(int)
